/*
 * Copyright (C) 2008-2009 Federal University of Pernambuco and 
 * University of Central Florida
 *
 * This file is part of AJML
 *
 * AJML is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * AJML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AJML; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: PreOrPostconditionMethod.java,v 1.0 2008/0715/19 9:32:03 henriquerebelo Exp $
 * 
 * This file is based on the original $Id: PreOrPostconditionMethod.java,v 1.13 2005/12/09 19:46:03 f_rioux Exp $
 * by Yoonsik Cheon
 */

package org.aspectjml.ajmlrac;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.aspectjml.checker.Constants;
import org.aspectjml.checker.JmlAbstractVisitor;
import org.aspectjml.checker.JmlConstructorDeclaration;
import org.aspectjml.checker.JmlFormalParameter;
import org.aspectjml.checker.JmlMessages;
import org.aspectjml.checker.JmlMethodDeclaration;
import org.aspectjml.checker.JmlTypeDeclaration;
import org.aspectjml.util.AspectUtil;
import org.aspectjml.util.QDoxUtil;
import org.multijava.mjc.CClassType;
import org.multijava.mjc.CMethod;
import org.multijava.mjc.CType;
import org.multijava.mjc.JFormalParameter;
import org.multijava.mjc.JStatement;
import org.multijava.util.compiler.TokenReference;

import com.thoughtworks.qdox.model.Annotation;
import com.thoughtworks.qdox.model.JavaField;
import com.thoughtworks.qdox.model.JavaMethod;

/**
 * An abstract class for generating <em>precondition</em> or
 * <em>postconditin checking method as an AspectJ advice</em> for the given methods.
 * The generated assertion method automatic inherit its superclass's
 * assertions. 
 *
 * <p>
 * The class implements a variant of the <em>Template Pattern</em>
 * [GoF95], prescribed in the class {@link AssertionMethod}.
 * </p>
 *
 * @see AssertionMethod
 *
 * @author Henrique Rebelo
 * @version $Revision: 1.0 $
 */

public abstract class PreOrPostconditionMethod extends AssertionMethod
{
	/** Construct a new <code>PreOrPostconditionMethod</code> object. */
	protected PreOrPostconditionMethod(JmlTypeDeclaration typeDecl,
			JmlMethodDeclaration mdecl,
			String stackMethod) { 
		this.typeDecl = typeDecl;
		this.methodDecl = mdecl;
		name = mdecl.ident();
		this.methodIdentification = "\"" + name + "\"";
		this.parameters = mdecl.parameters();

		boolean isMethodCrosscutSpecChecking = AspectUtil.getInstance().isCrosscutSpecChecking(this.methodDecl);
		if(isMethodCrosscutSpecChecking){
			JavaMethod crosscutSpecCheckingMethod = AspectUtil.getInstance().getJavaMethCrosscutSpecChecking(this.methodDecl);
			Annotation pcAnno = null;
        	// getting pointcut's value
        	for (int i = 0; i < crosscutSpecCheckingMethod.getAnnotations().length; i++) {
        		Annotation currentAnno = crosscutSpecCheckingMethod.getAnnotations()[i];
        		if((currentAnno.toString().startsWith("@org.aspectj.lang.annotation.Pointcut"))){
        			pcAnno = currentAnno;
        			break;
        		}
        	}
        	String xcsPointcut = pcAnno.getProperty("value").getParameterValue().toString().
        			replaceAll("\"(\\s)*\\+(\\s)*\"", "").replace("\"", "");
        	Pattern staticPattern = Pattern.compile("\\bstatic\\b");
    		Matcher staticMatcher = staticPattern.matcher(xcsPointcut.replaceAll("!(\\s)*static", ""));

    		if(staticMatcher.find()){
    			this.isStatic = true;
    		}
		}
		else{
			this.isStatic = this.methodDecl.isStatic();
		}
		this.exceptionToThrow = "JMLAssertionError";
	}

	// ----------------------------------------------------------------------
	// GENERATION
	// ----------------------------------------------------------------------

	/** Indicates whether the generated assertion method should try to
	 * dynamically inherit the corresponding assertion method of the
	 * superclass or interfaces.  Method specifications are inherited
	 * only for instance (non-static) methods; i.e., class (static)
	 * method specifications are not inherited by subclasses. Also,
	 * method specifications for constructors are not inherited by
	 * subclasses.  Dynamic inheritance is costly, so attempted only
	 * if the method has a superclass (or interfaces) and is
	 * overriding any of its superclass methods.
	 *
	 * <pre><jml>
	 * also
	 * protected normal_behavior
	 *   ensures \result == !isStatic 
	 *     && !(methodDecl instanceof JmlConstructorDeclaration)
	 *     && methodDecl.isOverriding();
	 * </jml></pre>
	 */
	protected /*@ pure @*/ boolean canInherit() {
		return 
		// don't try to inherit from Object
		(hasExplicitSuperClass() || typeDecl.interfaces().length > 0) 
		&& !isStatic
		&& !(methodDecl instanceof JmlConstructorDeclaration)
		&& methodDecl.isOverriding();
	}

	/**
	 * Returns the name of assertion check method to be generated for
	 * the given method declaration. For constructors, the result is 
	 * MN_INIT; for non-constructors, the result is the same as the name
	 * of the given method declaration.
	 */    
	protected static String methodName(JmlMethodDeclaration mdecl) {
		return (mdecl instanceof JmlConstructorDeclaration) ? 
				MN_INIT : mdecl.ident();
	}
	
//	protected boolean isLocalToCurrentType(){
//		
//	}

	/**
	 * @param code
	 */
	protected StringBuffer buildAdviceHeader(String assertionMethType, String instrumentationType, long visibility, 
			boolean isMethodCrosscutSpecChecking) {
		final StringBuffer code = new StringBuffer("");
		String codeStr = "";
		String classQualifiedName = this.methodDecl.getMethod().owner().getJavaName();
		String methodQualifiedName = null;
		String methodQualifiedNameJP = this.methodDecl.getMethod().getJavaName();
		
		if (this.methodDecl.isConstructor()){
			if(this.methodDecl.getMethod().owner().isInnerClass()){
				methodQualifiedName = this.methodDecl.stringRepresentation();
				methodQualifiedName = methodQualifiedName.replace(this.typeDecl.getCClass().getJavaName()+", ", "");
				methodQualifiedName = methodQualifiedName.replace(this.methodDecl.getMethod().owner().ident(), 
						this.methodDecl.getMethod().owner().ident()+".new").replaceAll("<(.)*>",""); // removing the Java 5 Diamond --- [[[hemr]]]
			}
			else{
				methodQualifiedName = this.methodDecl.stringRepresentation().replace(this.methodDecl.getMethod().owner().ident(), 
						this.methodDecl.getMethod().owner().ident()+".new").replaceAll("<(.)*>",""); // removing the Java 5 Diamond --- [[[hemr]]]
			}
		}
		else{
			if(this.methodDecl.isStatic()){
				methodQualifiedName = this.methodDecl.getMethod().getJavaName();	
			}	
			else{
				if(instrumentationType.equals("clientAwareChecking") || instrumentationType.equals("executionSite")) {
					methodQualifiedName = this.methodDecl.getMethod().getJavaName();	
				}
				else{
					methodQualifiedName = this.methodDecl.getMethod().getJavaName().replace(classQualifiedName+".", "");	
				}
			}
		}
		
		// handling crosscutting contract specifications --- [[[hemr]]]
        String xcsPointcut = "";
        String xcsReturningOrThrowProperty = "";
        boolean adviceinterception = true;
        boolean reentering = false;
        boolean selfreentering = false;
        boolean subtyping = true;
        
        boolean isPointcutMethStatic = false;
        boolean flexibleXCS = false;
        boolean isReviseMeth = false;
        boolean isAppendLikeMeth = false;
        boolean isReviseMethCallPC = false;
        boolean isAppendLikeMethCallPC = false;
		
        if(isMethodCrosscutSpecChecking){
        	JavaMethod crosscutSpecCheckingMethod = AspectUtil.getInstance().getJavaMethCrosscutSpecChecking(this.methodDecl);
        	String xcsParentClassName = "";
        	String xcsParentClassFullyQualifiedName = "";
        	
        	// Is the pointcut meth static ?
        	isPointcutMethStatic= crosscutSpecCheckingMethod.isStatic();

        	// checking if the crosscutSpecCheckingMethod is for an interface
    		if(AspectUtil.getInstance().isCrosscuttingContractSpecificationForInterface(crosscutSpecCheckingMethod)){
    			classQualifiedName = AspectUtil.replaceDollaSymbol(crosscutSpecCheckingMethod.getParentClass().getParentClass().getFullyQualifiedName());
        		xcsParentClassName = crosscutSpecCheckingMethod.getParentClass().getParentClass().getName();
        		xcsParentClassFullyQualifiedName = crosscutSpecCheckingMethod.getParentClass().getParentClass().getFullyQualifiedName().replace("$", ".");
    		}
    		else{
    			classQualifiedName = AspectUtil.replaceDollaSymbol(crosscutSpecCheckingMethod.getParentClass().getFullyQualifiedName());
        		xcsParentClassName = crosscutSpecCheckingMethod.getParentClass().getName();
        		xcsParentClassFullyQualifiedName = crosscutSpecCheckingMethod.getParentClass().getFullyQualifiedName().replace("$", ".");;
    		}
        	
        	Annotation pcAnno = null;
        	// getting pointcut's value
        	for (int i = 0; i < crosscutSpecCheckingMethod.getAnnotations().length; i++) {
        		Annotation currentAnno = crosscutSpecCheckingMethod.getAnnotations()[i];
        		if((currentAnno.toString().startsWith("@org.aspectj.lang.annotation.Pointcut"))){
        			pcAnno = currentAnno;
        			break;
        		}
        	}
        	
        	if(AspectUtil.getInstance().isTypeDeclAspectDecl(crosscutSpecCheckingMethod.getParentClass())){
        		if(AspectUtil.getInstance().isXCSFlexible(crosscutSpecCheckingMethod)){
            		flexibleXCS = true;
            	}
			}
        	else {
        		if(AspectUtil.getInstance().isXCSFlexible(crosscutSpecCheckingMethod)){
        			JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), pcAnno.getLineNumber()), 
							JmlMessages.INVALID_FLEXIBLE_OR_GLOBAL_POINTCUT_XCS, this.methodDecl.ident());
					System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
            	}
        	}
        	
        	String xcsPCTmp = pcAnno.getProperty("value").getParameterValue().toString().
        			replaceAll("\"(\\s)*\\+(\\s)*\"", "").replace("\"", "");
        	
        	// extrating a pc value through a constant string (if any) --- [[[hemr]]]
    		xcsPCTmp = this.extractPCValue(xcsPCTmp);
        	
        	xcsPCTmp = processingXCSPointcut(xcsParentClassName,
					xcsParentClassFullyQualifiedName, pcAnno, xcsPCTmp, flexibleXCS);
        	
        	if(!flexibleXCS){
        		// processing reusable pointcuts --- [[[hemr]]]
    			List javaMethXCS = AspectUtil.getInstance().getCrosscutSpecMeths();
    			for (Iterator iterator = javaMethXCS.iterator(); iterator
    					.hasNext();) {
    				JavaMethod currentJavaMethod = (JavaMethod) iterator.next();
    				
    				Pattern thisPattern = Pattern.compile("\\.\\b"+Pattern.quote(currentJavaMethod.getName())+"\\b");
    				Matcher thisMatcher = thisPattern.matcher(xcsPCTmp);
    				if(!thisMatcher.find()){
    					Pattern thisPatternAux = Pattern.compile("\\b"+Pattern.quote(currentJavaMethod.getName())+"\\b");
    					Matcher thisMatcherAux = thisPatternAux.matcher(xcsPCTmp);
    					if(thisMatcherAux.find()){
    						//testing if it is local
    						if(currentJavaMethod.getParentClass().getFullyQualifiedName().equals(typeDecl.getCClass().getJavaName())){
    				        	for (int i = 0; i < currentJavaMethod.getAnnotations().length; i++) {
    				        		Annotation currentAnno = currentJavaMethod.getAnnotations()[i];
    				        		if((currentAnno.toString().startsWith("@org.aspectj.lang.annotation.Pointcut"))){
    				        			pcAnno = currentAnno;
    				        			xcsPCTmp = xcsPCTmp.replace(thisMatcherAux.group(), typeDecl.getCClass().getJavaName()+"."+thisMatcherAux.group());
    				        			break;
    				        		}
    				        	}
    						}
    						else {
    							ArrayList allOverriddenMethodsFromSuperClasses = new ArrayList();
    							this.typeDecl.getCClass().getAllMethods(allOverriddenMethodsFromSuperClasses);
    							for (Iterator iterator2 = allOverriddenMethodsFromSuperClasses
    									.iterator(); iterator2.hasNext();) {
    								CMethod currentInheritedMeth = (CMethod) iterator2.next();
    								if(currentJavaMethod.getParentClass().getFullyQualifiedName().equals(currentInheritedMeth.owner().getJavaName())){
    						        	for (int i = 0; i < currentJavaMethod.getAnnotations().length; i++) {
    						        		Annotation currentAnno = currentJavaMethod.getAnnotations()[i];
    						        		if((currentAnno.toString().startsWith("@org.aspectj.lang.annotation.Pointcut"))){
    						        			pcAnno = currentAnno;
    						        			xcsPCTmp = xcsPCTmp.replace(thisMatcherAux.group(), currentInheritedMeth.owner().getJavaName()+"."+thisMatcherAux.group());
    						        			break;
    						        		}
    						        	}
    								}
    							}
    						}
    					}
    				}
    			}
    			
    			// avoiding property-based type pattern in XCS --- [[[hemr]]]
    			Pattern namePropertyBasedTypePattern = Pattern.compile("\\*(\\.)?(\\.)?\\b"+Pattern.quote(xcsParentClassFullyQualifiedName)+"\\b"+"(\\s)*(\\+)?(\\s)*\\.");
    			Matcher namePropertyBasedTypePatternMatcher = namePropertyBasedTypePattern.matcher(xcsPCTmp);
    			if(namePropertyBasedTypePatternMatcher.find()){
    				JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), pcAnno.getLineNumber()), 
    						JmlMessages.PROPERTY_BASED_POINTCUT_NOT_ALLOWED, this.methodDecl.ident(), xcsParentClassFullyQualifiedName);
    				System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
    			}
    			
    			namePropertyBasedTypePattern = Pattern.compile("\\b"+Pattern.quote(xcsParentClassFullyQualifiedName)+"\\b"+"(\\*)?(\\.)?(\\.)?\\*(\\s)*(\\+)?(\\s)*\\.");
    			namePropertyBasedTypePatternMatcher = namePropertyBasedTypePattern.matcher(xcsPCTmp);
    			if(namePropertyBasedTypePatternMatcher.find()){
    				JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), pcAnno.getLineNumber()), 
    						JmlMessages.PROPERTY_BASED_POINTCUT_NOT_ALLOWED, this.methodDecl.ident(), xcsParentClassFullyQualifiedName);
    				System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
    			}
    			
    			namePropertyBasedTypePattern = Pattern.compile("\\b"+Pattern.quote(xcsParentClassFullyQualifiedName)+"\\b"+"(\\*)?(\\.\\.)?(\\*(\\w)*(\\*)?)*?(\\s)*\\(");
    			namePropertyBasedTypePatternMatcher = namePropertyBasedTypePattern.matcher(xcsPCTmp);
    			if(namePropertyBasedTypePatternMatcher.find()){
    				JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), pcAnno.getLineNumber()), 
    						JmlMessages.PROPERTY_BASED_POINTCUT_NOT_ALLOWED, this.methodDecl.ident(), xcsParentClassFullyQualifiedName);
    				System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
    			}
    			
    			// handling static join points --- [[[hemr]]]
    			// #1 verifying if all static pointcuts if any are defined for all pointcuts
    			Pattern staticPattern = Pattern.compile("\\bstatic\\b");
        		Matcher staticMatcher = staticPattern.matcher(xcsPCTmp.replaceAll("!(\\s)*static\\b", ""));
        		int execCallCount = 0;
        		if(staticMatcher.find()){
        			int staticCount = 1;
        			isPointcutMethStatic = true;
    				while(staticMatcher.find()){
    					staticCount++;
    				}
    				
    				Pattern execCallPattern = Pattern.compile("(execution(\\s)*\\()|(call(\\s)*\\()");
    				Matcher execCallMatcher = execCallPattern.matcher(xcsPCTmp);
    				while(execCallMatcher.find()){
    					execCallCount++;
    				}
    				if(!(staticCount == execCallCount)){
    					if(!crosscutSpecCheckingMethod.isStatic()){
    						JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), pcAnno.getLineNumber()), 
        							JmlMessages.INVALID_STATIC_AND_INSTANCE_POINTCUT_TOGETHER, this.methodDecl.ident());
        					System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
    					}
    				}
        		}
        		else {
        			execCallCount = -1;
        		}
        		
        		// #1.2 verifying the use of adviceexecution(), which is not allowed in AspectJML
        		Pattern adviceexecutionPattern = Pattern.compile("\\badviceexecution\\b(\\s)*\\((\\s)*\\)");
        		Matcher adviceexecutionMatcher = adviceexecutionPattern.matcher(xcsPCTmp);
        		if(adviceexecutionMatcher.find()){
        			JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), pcAnno.getLineNumber()), 
							JmlMessages.INVALID_ADVICEEXECUTION_AND_POINTCUT, this.methodDecl.ident());
					System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
        		}
        		
        		// #2 if they are valid static pointcuts, we must ensure that they mention its enclosing type
        		if(isPointcutMethStatic){        			
        			// #3 static pointcuts cannot refer to instance members
        			if(AspectUtil.getInstance().hasThisRef){
        				JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), pcAnno.getLineNumber()), 
        						JmlMessages.NO_THIS_REF_IN_STATIC_POINTCUT, this.methodDecl.ident());
        				System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
        			}
        			
        			// util for code generation in XCS mode --- [[[hemr]]]
        			AspectUtil.getInstance().isStaticPC = true;
        		}
        	}
			
    		// getting the processed PC
        	xcsPointcut = xcsPCTmp;
//        	System.out.println("xcsPointcut = "+xcsPointcut);
        }
        else{
        	JavaMethod adviceMethod = AspectUtil.getInstance().getJavaMethAdviceCrosscutting(this.methodDecl);
        	
        	if(adviceMethod != null){
        		String xcsParentClassName = "";
        		String xcsParentClassFullyQualifiedName = "";

        		// checking if the crosscutSpecCheckingMethod is for an interface
        		if(AspectUtil.getInstance().isCrosscuttingContractSpecificationForInterface(adviceMethod)){
        			classQualifiedName = AspectUtil.replaceDollaSymbol(adviceMethod.getParentClass().getParentClass().getFullyQualifiedName());
        			xcsParentClassName = adviceMethod.getParentClass().getParentClass().getName();
        			xcsParentClassFullyQualifiedName = adviceMethod.getParentClass().getParentClass().getFullyQualifiedName().replace("$", ".");
        		}
        		else{
        			classQualifiedName = AspectUtil.replaceDollaSymbol(adviceMethod.getParentClass().getFullyQualifiedName());
        			xcsParentClassName = adviceMethod.getParentClass().getName();
        			xcsParentClassFullyQualifiedName = adviceMethod.getParentClass().getFullyQualifiedName().replace("$", ".");;
        		}

        		Annotation pcAnno = null;
        		
        		// getting pointcut's value
        		for (int i = 0; i < adviceMethod.getAnnotations().length; i++) {
        			Annotation currentAnno = adviceMethod.getAnnotations()[i];
        			if((currentAnno.toString().startsWith("@org.aspectj.lang.annotation.Pointcut"))){
        				if(pcAnno != null){
        					JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), currentAnno.getLineNumber()), 
        							JmlMessages.MIXING_ASPECT_ANNOTATION_NOT_ALLOWED, this.methodDecl.ident(), xcsParentClassFullyQualifiedName);
        					System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
        				}
        				pcAnno = currentAnno;
        			}
        			if((currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.Before"))){
        				if(pcAnno != null){
        					JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), currentAnno.getLineNumber()), 
        							JmlMessages.MIXING_ASPECT_ANNOTATION_NOT_ALLOWED, this.methodDecl.ident(), xcsParentClassFullyQualifiedName);
        					System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
        				}
        				pcAnno = currentAnno;
        			}
        			if((currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.After"))){
        				if(pcAnno != null){
        					JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), currentAnno.getLineNumber()), 
        							JmlMessages.MIXING_ASPECT_ANNOTATION_NOT_ALLOWED, this.methodDecl.ident(), xcsParentClassFullyQualifiedName);
        					System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
        				}
        				pcAnno = currentAnno;
        			}
        			if((currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.Around"))){
        				if(pcAnno != null){
        					JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), currentAnno.getLineNumber()), 
        							JmlMessages.MIXING_ASPECT_ANNOTATION_NOT_ALLOWED, this.methodDecl.ident(), xcsParentClassFullyQualifiedName);
        					System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
        				}
        				pcAnno = currentAnno;
        			}
        			if((currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.Revise"))){
        				if(pcAnno != null){
        					JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), currentAnno.getLineNumber()), 
        							JmlMessages.MIXING_ASPECT_ANNOTATION_NOT_ALLOWED, this.methodDecl.ident(), xcsParentClassFullyQualifiedName);
        					System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
        				}
        				pcAnno = currentAnno;
        				isReviseMeth = true;
        				if(pcAnno.getProperty("callsite") != null){
        					isReviseMethCallPC = new Boolean(pcAnno.getProperty("callsite").toString()).booleanValue();
            			}
        			}
        			if((currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.Append"))){
        				if(pcAnno != null){
        					JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), currentAnno.getLineNumber()), 
        							JmlMessages.MIXING_ASPECT_ANNOTATION_NOT_ALLOWED, this.methodDecl.ident(), xcsParentClassFullyQualifiedName);
        					System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
        				}
        				pcAnno = currentAnno;
        				isAppendLikeMeth = true;
        				if(pcAnno.getProperty("callsite") != null){
        					isAppendLikeMethCallPC = new Boolean(pcAnno.getProperty("callsite").toString()).booleanValue();
            			}
        			}
        			if((currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.Add"))){
        				if(pcAnno != null){
        					JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), currentAnno.getLineNumber()), 
        							JmlMessages.MIXING_ASPECT_ANNOTATION_NOT_ALLOWED, this.methodDecl.ident(), xcsParentClassFullyQualifiedName);
        					System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
        				}
        				pcAnno = currentAnno;
        				isAppendLikeMeth = true;
        			}
        			
        			if(AspectUtil.getInstance().isTypeDeclAspectDecl(adviceMethod.getParentClass())){
        				flexibleXCS = true;
//        				isPointcutMethStatic = true;
        			}
        			
        			if((AspectUtil.getInstance().isXCSFlexible(adviceMethod))){
        				JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), adviceMethod.getLineNumber()), 
        						JmlMessages.INVALID_FLEXIBLE_OR_GLOBAL_ADVICE, this.methodDecl.ident());
        				System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
        			}
        		}
        		String xcsPCTmp = "";
        		
        		if(!isReviseMeth && !isAppendLikeMeth){
        			if(pcAnno.getProperty("value") != null){
            			xcsPCTmp = pcAnno.getProperty("value").getParameterValue().toString().
                				replaceAll("\"(\\s)*\\+(\\s)*\"", "").replace("\"", "");
            			// this is removed... need to be sure, but does match Java semantics --- [[[hemr]]]
            			// ensuring the semantics of Java
//        				if(adviceMethod.isStatic()){
//        					subtyping = false;
//        				}
            		}
            		else{
            			xcsPCTmp = pcAnno.getProperty("pointcut").getParameterValue().toString().
                				replaceAll("\"(\\s)*\\+(\\s)*\"", "").replace("\"", "");
            			
            			if(pcAnno.getProperty("returning") != null){
            				xcsReturningOrThrowProperty = "returning="+pcAnno.getProperty("returning").toString();
            			}
            			else if(pcAnno.getProperty("throwing") != null){
            				xcsReturningOrThrowProperty = "throwing="+pcAnno.getProperty("throwing").toString();
            			}
            			
            			if(pcAnno.getProperty("adviceinterception") != null){
            				adviceinterception = new Boolean(pcAnno.getProperty("adviceinterception").toString()).booleanValue();
            			}
            			if(pcAnno.getProperty("reentering") != null){
            				reentering = new Boolean(pcAnno.getProperty("reentering").toString()).booleanValue();
            			}
            			if(pcAnno.getProperty("selfreentering") != null){
            				selfreentering = new Boolean(pcAnno.getProperty("selfreentering").toString()).booleanValue();
            			}
            			if(pcAnno.getProperty("subtyping") != null){
            				subtyping = new Boolean(pcAnno.getProperty("subtyping").toString()).booleanValue();
            			}
            			// this is removed... need to be sure, but does match Java semantics --- [[[hemr]]]
//            			else{
//            				// ensuring the semantics of Java
//            				if(adviceMethod.isStatic()){
//            					subtyping = false;
//            				}
//            			}
            		}
        		}
        		
        		// extrating a pc value through a constant string (if any) --- [[[hemr]]]
        		xcsPCTmp = this.extractPCValue(xcsPCTmp);

        		xcsPCTmp = processingXCSPointcut(xcsParentClassName,
        				xcsParentClassFullyQualifiedName, pcAnno, xcsPCTmp, flexibleXCS);
        		
        		// handling optional args pointcut :-) --- [[[hemr]]]
        		Pattern argsPattern = Pattern.compile("(&&(\\s)*\\bargs\\b\\((.)+\\))");
        		Matcher argsMatcher = argsPattern.matcher(xcsPCTmp);
        		if(!argsMatcher.find()){
        			StringBuffer args = new StringBuffer("");
        			this.generateOptionalArgsPC(parameters, args);
        			if(!args.toString().equals(" && args(..,..)")){
//        				System.out.println("before = "+args.toString());
//        				System.out.println("after = "+args.toString().replace("..,", "").replace(", ..", ""));
        				xcsPCTmp = "(" + xcsPCTmp + ")" + args.toString().replace("..,", "").replace(", ..", "");
//        				xcsPCTmp = "(" + xcsPCTmp + ")" + args.toString();
        			}
        		}

        		if(!flexibleXCS && !isReviseMeth){
        			// processing reusable pointcuts --- [[[hemr]]]
        			List javaMethXCS = AspectUtil.getInstance().getCrosscutSpecMeths();
        			for (Iterator iterator = javaMethXCS.iterator(); iterator
        					.hasNext();) {
        				JavaMethod currentJavaMethod = (JavaMethod) iterator.next();

        				Pattern thisPattern = Pattern.compile("\\.\\b"+Pattern.quote(currentJavaMethod.getName())+"\\b");
        				Matcher thisMatcher = thisPattern.matcher(xcsPCTmp);
        				if(!thisMatcher.find()){
        					Pattern thisPatternAux = Pattern.compile("\\b"+Pattern.quote(currentJavaMethod.getName())+"\\b");
        					Matcher thisMatcherAux = thisPatternAux.matcher(xcsPCTmp);
        					if(thisMatcherAux.find()){
        						//testing if it is local
        						if(currentJavaMethod.getParentClass().getFullyQualifiedName().equals(typeDecl.getCClass().getJavaName())){
        							xcsPCTmp = xcsPCTmp.replace(thisMatcherAux.group(), typeDecl.getCClass().getJavaName()+"."+thisMatcherAux.group());
        							// verifying if the reused pointcut is static
        							for (int i = 0; i < currentJavaMethod.getAnnotations().length; i++) {
        								Annotation currentAnno = currentJavaMethod.getAnnotations()[i];
        								if((currentAnno.toString().startsWith("@org.aspectj.lang.annotation.Pointcut"))){
        									pcAnno = currentAnno;
        									break;
        								}
        							}
        						}
        						else {
        							ArrayList allOverriddenMethodsFromSuperClasses = new ArrayList();
        							this.typeDecl.getCClass().getAllMethods(allOverriddenMethodsFromSuperClasses);
        							for (Iterator iterator2 = allOverriddenMethodsFromSuperClasses
        									.iterator(); iterator2.hasNext();) {
        								CMethod currentInheritedMeth = (CMethod) iterator2.next();
        								if(currentJavaMethod.getParentClass().getFullyQualifiedName().equals(currentInheritedMeth.owner().getJavaName())){
        									xcsPCTmp = xcsPCTmp.replace(thisMatcherAux.group(), currentInheritedMeth.owner().getJavaName()+"."+thisMatcherAux.group());
        									// verifying if the reused pointcut is static
        									for (int i = 0; i < currentJavaMethod.getAnnotations().length; i++) {
        										Annotation currentAnno = currentJavaMethod.getAnnotations()[i];
        										if((currentAnno.toString().startsWith("@org.aspectj.lang.annotation.Pointcut"))){
        											pcAnno = currentAnno;
        											break;
        										}
        									}
        								}
        							}
        						}
        					}
        				}
        			}

        			// avoiding property-based type pattern in XCS --- [[[hemr]]]
        			Pattern namePropertyBasedTypePattern = Pattern.compile("\\*(\\.)?(\\.)?\\b"+Pattern.quote(xcsParentClassFullyQualifiedName)+"\\b"+"(\\s)*(\\+)?(\\s)*\\.");
        			Matcher namePropertyBasedTypePatternMatcher = namePropertyBasedTypePattern.matcher(xcsPCTmp);
        			if(namePropertyBasedTypePatternMatcher.find()){
        				JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), pcAnno.getLineNumber()), 
        						JmlMessages.PROPERTY_BASED_POINTCUT_NOT_ALLOWED, this.methodDecl.ident(), xcsParentClassFullyQualifiedName);
        				System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
        			}

        			namePropertyBasedTypePattern = Pattern.compile("\\b"+Pattern.quote(xcsParentClassFullyQualifiedName)+"\\b"+"(\\*)?(\\.)?(\\.)?\\*(\\s)*(\\+)?(\\s)*\\.");
        			namePropertyBasedTypePatternMatcher = namePropertyBasedTypePattern.matcher(xcsPCTmp);
        			if(namePropertyBasedTypePatternMatcher.find()){
        				JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), pcAnno.getLineNumber()), 
        						JmlMessages.PROPERTY_BASED_POINTCUT_NOT_ALLOWED, this.methodDecl.ident(), xcsParentClassFullyQualifiedName);
        				System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
        			}

        			namePropertyBasedTypePattern = Pattern.compile("\\b"+Pattern.quote(xcsParentClassFullyQualifiedName)+"\\b"+"(\\*)?(\\.\\.)?(\\*(\\w)*(\\*)?)*?(\\s)*\\(");
        			namePropertyBasedTypePatternMatcher = namePropertyBasedTypePattern.matcher(xcsPCTmp);
        			if(namePropertyBasedTypePatternMatcher.find()){
        				JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), pcAnno.getLineNumber()), 
        						JmlMessages.PROPERTY_BASED_POINTCUT_NOT_ALLOWED, this.methodDecl.ident(), xcsParentClassFullyQualifiedName);
        				System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
        			}

        			// handling static join points --- [[[hemr]]]
        			// #1 verifying if all static pointcuts if any are defined for all pointcuts
        			Pattern staticPattern = Pattern.compile("\\bstatic\\b");
        			Matcher staticMatcher = staticPattern.matcher(xcsPCTmp.replaceAll("!(\\s)*static\\b", ""));
        			int execCallCount = 0;
        			if(staticMatcher.find()){
        				int staticCount = 1;
        				isPointcutMethStatic = true;
        				while(staticMatcher.find()){
        					staticCount++;
        				}

        				Pattern execCallPattern = Pattern.compile("(execution(\\s)*\\()|(call(\\s)*\\()|(set(\\s)*\\))|(get(\\s)*\\)|(withincode(\\s)*\\())");
        				Matcher execCallMatcher = execCallPattern.matcher(xcsPCTmp);
        				while(execCallMatcher.find()){
        					execCallCount++;
        				}
        				if(!(staticCount == execCallCount)){
        					if(!adviceMethod.isStatic()){
        						JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), pcAnno.getLineNumber()), 
            							JmlMessages.INVALID_STATIC_AND_INSTANCE_POINTCUT_TOGETHER, this.methodDecl.ident());
            					System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
        					}
        				}
        			}
        			else {
        				execCallCount = -1;
        			}
        			
        			// #1.2 verifying the use of adviceexecution(), which is not allowed in AspectJML
            		Pattern adviceexecutionPattern = Pattern.compile("\\badviceexecution\\b(\\s)*\\((\\s)*\\)");
            		Matcher adviceexecutionMatcher = adviceexecutionPattern.matcher(xcsPCTmp);
            		if(adviceexecutionMatcher.find()){
            			JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), pcAnno.getLineNumber()), 
    							JmlMessages.INVALID_ADVICEEXECUTION_AND_ADVICE, this.methodDecl.ident());
    					System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
            		}

        			// #2 if they are valid static pointcuts, we must ensure that they mention its enclosing type
            		if(isPointcutMethStatic){
            			// #3 static pointcuts cannot refer to instance members
            			if(AspectUtil.getInstance().hasThisRef){
            				JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), pcAnno.getLineNumber()), 
            						JmlMessages.NO_THIS_REF_IN_STATIC_POINTCUT, this.methodDecl.ident());
            				System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
            			}

            			// util for code generation in XCS mode --- [[[hemr]]]
            			AspectUtil.getInstance().isStaticPC = true;
            		}
        		}

        		// getting the processed PC
        		xcsPointcut = xcsPCTmp;
//        		System.out.println("Method = "+this.methodDecl.ident());
//        		System.out.println("xcsPointcut = "+xcsPointcut);
//        		System.out.println();
        	}
        	
        	//********************************************************************
        	classQualifiedName = AspectUtil.replaceDollaSymbol(classQualifiedName);
        }
		if (assertionMethType.equals("PreconditionAssertionMethod")){
			codeStr = this.buildPreconditionAdviceHeader(classQualifiedName, methodQualifiedName, 
					instrumentationType, visibility, methodQualifiedNameJP, isMethodCrosscutSpecChecking, xcsPointcut, isPointcutMethStatic, flexibleXCS);	
		}
		else if (assertionMethType.equals("NPAndXPAssertionMethodsWithAroundAdvice")){
			codeStr = this.buildNPAndXPAroundAdviceHeader(this.methodDecl.parameters(), classQualifiedName, methodQualifiedName, 
					instrumentationType, visibility, methodQualifiedNameJP, isMethodCrosscutSpecChecking, xcsPointcut, isPointcutMethStatic, flexibleXCS);
		}
		else if (assertionMethType.equals("NPAssertionMethodWithAfterReturningAdvice")){
			codeStr = this.buildNPAfterReturningAdviceHeader(this.methodDecl.parameters(), classQualifiedName, methodQualifiedName, 
					instrumentationType, visibility, isMethodCrosscutSpecChecking, xcsPointcut, isPointcutMethStatic, flexibleXCS);
		}
		else if (assertionMethType.equals("XPAssertionMethodWithAfterThrowingAdvice")){
			codeStr = this.buildXPAfterThrowingAdviceHeader(this.methodDecl.parameters(), classQualifiedName, methodQualifiedName, 
					instrumentationType, visibility, isMethodCrosscutSpecChecking, xcsPointcut, isPointcutMethStatic, flexibleXCS);
		}
		code.append(AspectUtil.processMethSig(codeStr));
		
		// handling advising/crosscutting code in AspectJML --- [[[hemr]]]
		JavaMethod advicemethod = AspectUtil.getInstance().getJavaMethAdviceCrosscutting(this.methodDecl);
		if(advicemethod != null){
			// removing the used advicemethod --- [[[hemr]]]
			AspectUtil.getInstance().getAdviceDeclJavaFile().remove(advicemethod);
			// advice must be declared public --- [[[hemr]]]
			
//			if(!advicemethod.isPublic() && !isReviseMeth){
//				JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), advicemethod.getLineNumber()), 
//						JmlMessages.INVALID_ADVICE_DECLARATION_NON_PUBLIC, this.methodDecl.ident());
//				System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
//			}
			if(advicemethod.isConstructor() && !isReviseMeth && !isAppendLikeMeth){
				JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), advicemethod.getLineNumber()), 
						JmlMessages.INVALID_ADVICE_DECLARATION_AS_CONSTRUCTOR, this.methodDecl.ident());
				System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
			}
			
			if(!AspectUtil.getInstance().isTypeDeclAspectDecl(advicemethod.getParentClass())){
				if((!advicemethod.isStatic()) && (isPointcutMethStatic)){
					JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), advicemethod.getLineNumber()), 
							JmlMessages.INVALID_INSTANCE_ADVICE_DECLARATION_WITH_STATIC_POINTCUT, this.methodDecl.ident());
					System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
				}	
			}
			else {
				if((advicemethod.isStatic())){
					JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), advicemethod.getLineNumber()), 
							JmlMessages.INVALID_ADVICE_DECLARATION_STATIC, this.methodDecl.ident());
					System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
				}
			}
			
			StringBuffer adviceCode = new StringBuffer("");
			boolean isAroundAdvice = false;
			for (int i = 0; i < advicemethod.getAnnotations().length; i++) {
    			Annotation currentAnno = advicemethod.getAnnotations()[i];
    			if((currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.Before"))){
    				if(!advicemethod.isConstructor() && ((advicemethod.getReturnType()!= null) && !advicemethod.getReturnType().getFullyQualifiedName().equals("void"))){
    					JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), advicemethod.getLineNumber()), 
    							JmlMessages.INVALID_ADVICE_DECLARATION_NON_VOID, "before", this.methodDecl.ident());
    					System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
    					
    				}
    				if(QDoxUtil.isAnnotatedWithXCS(advicemethod)){
    					JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), advicemethod.getLineNumber()), 
    							JmlMessages.INVALID_ADVICE_DECLARATION_AS_XCS, "before", this.methodDecl.ident());
    					System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
    				}
    				codeStr = this.buildBeforeAdviceHeaderForCC(classQualifiedName, xcsPointcut, advicemethod.isStatic(), false, isAppendLikeMethCallPC, adviceinterception, reentering, selfreentering);
    				adviceCode.append(codeStr);
    				break;
    			}
    			else if((currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.AfterReturning"))){
    				if(!advicemethod.isConstructor() && ((advicemethod.getReturnType()!= null) && !advicemethod.getReturnType().getFullyQualifiedName().equals("void"))){
    					JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), advicemethod.getLineNumber()), 
    							JmlMessages.INVALID_ADVICE_DECLARATION_NON_VOID, "after-returning", this.methodDecl.ident());
    					System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
    					
    				}
    				if(QDoxUtil.isAnnotatedWithXCS(advicemethod)){
    					JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), advicemethod.getLineNumber()), 
    							JmlMessages.INVALID_ADVICE_DECLARATION_AS_XCS, "after-returning", this.methodDecl.ident());
    					System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
    				}
    				codeStr = this.buildAfterReturningAdviceHeaderForCC(this.methodDecl.parameters(), classQualifiedName, xcsPointcut, 
    						xcsReturningOrThrowProperty, advicemethod.isStatic(), false, isAppendLikeMethCallPC, adviceinterception, reentering, selfreentering);
    				adviceCode.append(codeStr);
    				break;
    			}
    			else if((currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.AfterThrowing"))){
    				if(!advicemethod.isConstructor() && ((advicemethod.getReturnType()!= null) && !advicemethod.getReturnType().getFullyQualifiedName().equals("void"))){
    					JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), advicemethod.getLineNumber()), 
    							JmlMessages.INVALID_ADVICE_DECLARATION_NON_VOID, "after-throwing", this.methodDecl.ident());
    					System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
    					
    				}
    				if(QDoxUtil.isAnnotatedWithXCS(advicemethod)){
    					JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), advicemethod.getLineNumber()), 
    							JmlMessages.INVALID_ADVICE_DECLARATION_AS_XCS, "after-throwing", this.methodDecl.ident());
    					System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
    				}
    				codeStr = this.buildAfterThrowingAdviceHeaderForCC(this.methodDecl.parameters(), classQualifiedName, xcsPointcut, 
    						xcsReturningOrThrowProperty, advicemethod.isStatic(), false, isAppendLikeMethCallPC, adviceinterception, reentering, selfreentering);
    				adviceCode.append(codeStr);
    				break;
    			}
    			else if((currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.After"))){
    				if(!advicemethod.isConstructor() && ((advicemethod.getReturnType()!= null) && !advicemethod.getReturnType().getFullyQualifiedName().equals("void"))){
    					JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), advicemethod.getLineNumber()), 
    							JmlMessages.INVALID_ADVICE_DECLARATION_NON_VOID, "after", this.methodDecl.ident());
    					System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
    					
    				}
    				if(QDoxUtil.isAnnotatedWithXCS(advicemethod)){
    					JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), advicemethod.getLineNumber()), 
    							JmlMessages.INVALID_ADVICE_DECLARATION_AS_XCS, "after", this.methodDecl.ident());
    					System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
    				}
    				codeStr = this.buildAfterAdviceHeaderForCC(this.methodDecl.parameters(), classQualifiedName, methodQualifiedName, 
    						xcsPointcut, advicemethod.isStatic(), false, isAppendLikeMethCallPC, adviceinterception, reentering, selfreentering);
    				adviceCode.append(codeStr);
    				break;
    			}
    			else if((currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.Around"))){
    				isAroundAdvice = true;
    				codeStr = this.buildAroundAdviceHeaderForCC(this.methodDecl.parameters(), classQualifiedName, xcsPointcut, advicemethod.isStatic(), false, isReviseMethCallPC, adviceinterception, reentering, selfreentering);
    				adviceCode.append(codeStr);
    				break;
    			}
    			else if((currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.Revise"))){
    				codeStr = this.buildAroundAdviceHeaderForCC(this.methodDecl.parameters(), classQualifiedName, xcsPointcut, advicemethod.isStatic(), true, isReviseMethCallPC, adviceinterception, reentering, selfreentering);
    				adviceCode.append(codeStr);
    				break;
    			}
    			else if((currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.AddBefore"))){
    				codeStr = this.buildBeforeAdviceHeaderForCC(classQualifiedName, xcsPointcut, advicemethod.isStatic(), true, isAppendLikeMethCallPC, adviceinterception, reentering, selfreentering);
    				adviceCode.append(codeStr);
    				break;
    			}
    			else if((currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.AppendBefore"))){
    				codeStr = this.buildBeforeAdviceHeaderForCC(classQualifiedName, xcsPointcut, advicemethod.isStatic(), true, isAppendLikeMethCallPC, adviceinterception, reentering, selfreentering);
    				adviceCode.append(codeStr);
    				break;
    			}
    			else if((currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.AddAfterReturning"))){
    				codeStr = this.buildAfterReturningAdviceHeaderForCC(this.methodDecl.parameters(), classQualifiedName, xcsPointcut, 
    						xcsReturningOrThrowProperty, advicemethod.isStatic(), true, isAppendLikeMethCallPC, adviceinterception, reentering, selfreentering);
    				adviceCode.append(codeStr);
    				break;
    			}
    			else if((currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.AppendAfterReturning"))){
    				codeStr = this.buildAfterReturningAdviceHeaderForCC(this.methodDecl.parameters(), classQualifiedName, xcsPointcut, 
    						xcsReturningOrThrowProperty, advicemethod.isStatic(), true, isAppendLikeMethCallPC, adviceinterception, reentering, selfreentering);
    				adviceCode.append(codeStr);
    				break;
    			}
    			else if((currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.AddAfterThrowing"))){
    				codeStr = this.buildAfterThrowingAdviceHeaderForCC(this.methodDecl.parameters(), classQualifiedName, xcsPointcut, 
    						xcsReturningOrThrowProperty, advicemethod.isStatic(), true, isAppendLikeMethCallPC, adviceinterception, reentering, selfreentering);
    				adviceCode.append(codeStr);
    				break;
    			}
    			else if((currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.AppendAfterThrowing"))){
    				codeStr = this.buildAfterThrowingAdviceHeaderForCC(this.methodDecl.parameters(), classQualifiedName, xcsPointcut, 
    						xcsReturningOrThrowProperty, advicemethod.isStatic(), true, isAppendLikeMethCallPC, adviceinterception, reentering, selfreentering);
    				adviceCode.append(codeStr);
    				break;
    			}
    			else if((currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.AddAfter"))){
    				codeStr = this.buildAfterAdviceHeaderForCC(this.methodDecl.parameters(), classQualifiedName, methodQualifiedName, 
    						xcsPointcut, advicemethod.isStatic(), true, isAppendLikeMethCallPC, adviceinterception, reentering, selfreentering);
    				adviceCode.append(codeStr);
    				break;
    			}
    			else if((currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.AppendAfter"))){
    				codeStr = this.buildAfterAdviceHeaderForCC(this.methodDecl.parameters(), classQualifiedName, methodQualifiedName, 
    						xcsPointcut, advicemethod.isStatic(), true, isAppendLikeMethCallPC, adviceinterception, reentering, selfreentering);
    				adviceCode.append(codeStr);
    				break;
    			}
			}
			this.buildAdviceBody(codeStr, classQualifiedName, isReviseMeth, advicemethod, adviceCode, 
					isAroundAdvice, isAppendLikeMeth, subtyping, adviceinterception, xcsPointcut);
			this.saveAdviceDeclForCodeGeneration(advicemethod, adviceCode);
		}
		// validating declare soft on java methods --- [[[hemr]]]
		JavaMethod declareSoftMeth = AspectUtil.getInstance().getJavaMethDeclareSoftCrosscutting(this.methodDecl);
		if(declareSoftMeth != null){
			for (int i = 0; i < declareSoftMeth.getAnnotations().length; i++) {
    			Annotation currentAnno = declareSoftMeth.getAnnotations()[i];
    			if((currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.DeclareSoft"))){
    				if(currentAnno.getProperty("value") == null){
    					JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), currentAnno.getLineNumber()), 
    							JmlMessages.DECLARE_SOFT_ANNOTATION_REQUIRED_VALUE, this.methodDecl.ident(), 
    							classQualifiedName);
    					System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
    				}
    				
    				String declareSoft = "";
    				String anno = currentAnno.getProperty("value").getParameterValue().toString().
            				replaceAll("\"(\\s)*\\+(\\s)*\"", "").replace("\"", "");
    				if(declareSoftMeth.isConstructor()){
    					declareSoft = "declare soft: "+anno+": execution("+classQualifiedName+
    							".new"+AspectUtil.generateMethodParameters(this.methodDecl.parameters(), false)+");";
    				}
    				else{
    					declareSoft = "declare soft: "+anno+": execution(* "+
    							classQualifiedName+
    							"."+this.methodDecl.ident()+AspectUtil.generateMethodParameters(this.methodDecl.parameters(), false)+");";
    				}
    				String declSoft = declareSoft+
    						getDeclareSoftAnnotationMappingLocationInfo(declareSoftMeth, currentAnno);
    				if(canIncludeDeclareSoftMethodDecl(declSoft)){
    					AspectUtil.getInstance().appendDeclareSoftMethodDeclStmt(declSoft);
    				}
    				break;
    			}
			}
		}
		return code;
	}
	
	private boolean canIncludeDeclareSoftMethodDecl(String declSoft){
		boolean result = true;
		for (Iterator iterator = AspectUtil.getInstance().getDeclareSoftMethodStmtDecls().iterator(); iterator.hasNext();) {
			String currentDeclSoft = (String) iterator.next();
			if(currentDeclSoft.equals(declSoft)){
				result = false;
				break;
			}
		}
		
		return result;
	}
	
	private String getDeclareSoftAnnotationMappingLocationInfo(JavaMethod declareSoftMeth, Annotation currentAnno){	
		StringBuffer mappingInfo = new StringBuffer("");
		mappingInfo.append("//#");
		mappingInfo.append("'");
		mappingInfo.append(declareSoftMeth.getParentClass().getFullyQualifiedName().replace("$", "."));
		mappingInfo.append("'");
		mappingInfo.append(" ");
		mappingInfo.append("(");
		mappingInfo.append(declareSoftMeth.getSource().getFile().getName());
		mappingInfo.append(":");
		mappingInfo.append(currentAnno.getLineNumber());
		mappingInfo.append(")");
		return mappingInfo.toString();
	}

	private void saveAdviceDeclForCodeGeneration(JavaMethod advicemethod, StringBuffer adviceCode) {
		// include or not...
		boolean canAdd = true;
		for (Iterator iterator = AspectUtil.getInstance().getAdviceDeclForXActivation().iterator(); iterator.hasNext();) {
			String adviceDecl = (String) iterator.next();
			if(adviceDecl.equals(adviceCode.toString())){
				canAdd = false;
				break;
			}
		}
		if(canAdd){
			AspectUtil.getInstance().appendAdviceDeclForXActivation(adviceCode.toString());	
			// saving directly to the owner type
			advicemethod.getParentClass().appendAdviceDecl(adviceCode.toString());
			
			// preparing annotation @org.aspectjml.lang.annotation.Advice for advice methods declared in classes or interfaces in AspectJML --- [[[hemr]]]
			String adviceFullySignatureWithOwnerName = AspectUtil.getInstance().getFullyJavaMethodSignatureWithOwnerName(advicemethod);
			if(advicemethod.isConstructor()){
				AspectUtil.getInstance().appendDeclareMethodAnnoForAdvice("declare @method: "+adviceFullySignatureWithOwnerName+": @org.aspectjml.lang.annotation.Advice;");	
			}
			else{
				AspectUtil.getInstance().appendDeclareMethodAnnoForAdvice("declare @method: "+adviceFullySignatureWithOwnerName+": @org.aspectjml.lang.annotation.Advice;");	
			}
		}
	}

	private void buildAdviceBody(String codeStr, String classQualifiedName, boolean isReviseMeth,
			JavaMethod advicemethod, StringBuffer adviceCode, boolean isAroundAdvice, boolean isAppendLikeMeth, 
			boolean subtyping, boolean adviceinterception, String xcsPointuct) {
		adviceCode.append(" {");
		StringBuffer adviceMethModifiers = new StringBuffer("");
		// modifiers
		String [] modifiers = advicemethod.getModifiers();
		for (int i = 0; i < modifiers.length; i++) {
			String currentModifier = modifiers[i];
			adviceMethModifiers.append(currentModifier).append(" ");
		}
		boolean hasCflowTest = false;
		Pattern cflowPattern = Pattern.compile("(\\bcflow\\b)|(\\bcflowbelow\\b)|(\\bif\\b)");
		Matcher cflowMatcher = cflowPattern.matcher(xcsPointuct);
		if(cflowMatcher.find()){
			hasCflowTest = true;
		}
//		System.out.println("xcsPointuct = "+xcsPointuct);
//		System.out.println("hasCflowTest = "+hasCflowTest);
		
//		Mapping code for crossrefs --- [[[hemr]]]
		if(advicemethod.isConstructor()) {
			adviceCode.append("//#").append(adviceMethModifiers.toString()).append(" ").append(classQualifiedName+"."+advicemethod.getName()+QDoxUtil.getSignatureErasure(AspectUtil.generateMethodParameters(advicemethod.getParameters(), false).toString(), advicemethod));
		}
		else {
			adviceCode.append("//#").append(adviceMethModifiers.toString()).append(advicemethod.getReturnType().getFullyQualifiedName()).append(" ").append(classQualifiedName+"."+advicemethod.getName()+QDoxUtil.getSignatureErasure(AspectUtil.generateMethodParameters(advicemethod.getParameters(), false).toString(), advicemethod));
		}
		adviceCode.append("#").append("'").append(classQualifiedName).append("'").append(" ").
		append("(").append(advicemethod.getSource().getFile().getName()).append(":").append(this.methodDecl.getTokenReference().line()).append(")").
		append(hasCflowTest?"#[with runtime test]":"").append(isReviseMeth?"#[revise]":"");
		adviceCode.append("\n");
		adviceCode.append("    ");
		
		if(AspectUtil.getInstance().isTypeDeclAspectDecl(advicemethod.getParentClass())){
//				String originalCall = "object$rac."+advicemethod.getCallSignature().replaceAll("\\b"+VN_ASPECTJ_THISJOINPOINT+"\\b", VN_ASPECTJ_THISJOINPOINT+"_ajml");
			adviceCode.append(advicemethod.getSourceCode().trim().replaceAll("\\b"+VN_ASPECTJ_THISJOINPOINT+"\\b", VN_ASPECTJ_THISJOINPOINT+"_ajml")).append("\n");
		}
		else{
			String oldName = "";
			String newName = "";
			if(advicemethod.isStatic()){
				// checking if it is an around or not...
				if(isAroundAdvice || isReviseMeth){
					String pjpVarNAme = getProceedingJPVarDeclName(this.methodDecl.parameters());
					adviceCode.append("if(JMLChecker.checkJP($, true, \"#\", &, @, *))\n ".replace("$", pjpVarNAme).replace("#", classQualifiedName).
							replace("&", Boolean.toString(subtyping)).replace("@", Boolean.toString(adviceinterception)).replace("*", Boolean.toString(Main.aRacOptions.clientAwareChecking())));
				}
				else{
					String thisJPVarNAme = getThisJPVarDeclName(this.methodDecl.parameters());
					adviceCode.append("if(JMLChecker.checkJP($, true, \"#\", &, @, *))\n ".replace("$", thisJPVarNAme).replace("#", classQualifiedName).
							replace("&", Boolean.toString(subtyping)).replace("@", Boolean.toString(adviceinterception)).replace("*", Boolean.toString(Main.aRacOptions.clientAwareChecking())));
				}
				adviceCode.append("    ");
			}
			else {
				if(isAroundAdvice || isReviseMeth){
					String pjpVarNAme = getProceedingJPVarDeclName(this.methodDecl.parameters());
					adviceCode.append("if(JMLChecker.checkJP($, false, \"#\", &, @, *))\n ".replace("$", pjpVarNAme).replace("#", classQualifiedName).
							replace("&", Boolean.toString(subtyping)).replace("@", Boolean.toString(adviceinterception)).replace("*", Boolean.toString(Main.aRacOptions.clientAwareChecking())));
				}
				else{
					String thisJPVarNAme = getThisJPVarDeclName(this.methodDecl.parameters());
					adviceCode.append("if(JMLChecker.checkJP($, false, \"#\", &, @, *))\n ".replace("$", thisJPVarNAme).replace("#", classQualifiedName).
							replace("&", Boolean.toString(subtyping)).replace("@", Boolean.toString(adviceinterception)).replace("*", Boolean.toString(Main.aRacOptions.clientAwareChecking())));
				}
				adviceCode.append("    ");
			}
			if(!isAppendLikeMeth){
				if((advicemethod.getReturnType()!= null) && !advicemethod.getReturnType().getFullyQualifiedName().equals("void")){
					adviceCode.append("return ");
				}
			}
			oldName = advicemethod.getName();
			if(isReviseMeth){
				
				if(advicemethod.isConstructor()){
					newName = "new$Revise";
				}
				else{
					newName = advicemethod.getName()+"$Revise";
				}
				newName += "$"+advicemethod.getSource().getFile().getName().replace(".", "_");
				advicemethod.setName(newName);
			}
			else if(isAppendLikeMeth){
				if(AspectUtil.getInstance().isAspectJMLAppendBefore(advicemethod)){
					if(advicemethod.isConstructor()){
						newName = "new$AppendBefore";
					}
					else{
						newName = advicemethod.getName()+"$AppendBefore";
					}
				}
				else if(AspectUtil.getInstance().isAspectJMLAppendAfter(advicemethod)){
					if(advicemethod.isConstructor()){
						newName = "new$AppendAfter";
					}
					else{
						newName = advicemethod.getName()+"$AppendAfter";
					}
				}
				else if(AspectUtil.getInstance().isAspectJMLAppendAfterReturning(advicemethod)){
					if(advicemethod.isConstructor()){
						newName = "new$AppendAfterReturning";
					}
					else{
						newName = advicemethod.getName()+"$AppendAfterReturning";
					}
				}
				else if(AspectUtil.getInstance().isAspectJMLAppendAfterThrowing(advicemethod)){
					if(advicemethod.isConstructor()){
						newName = "new$AppendAfterThrowing";
					}
					else{
						newName = advicemethod.getName()+"$AppendAfterThrowing";
					}
				}
				newName += "$"+advicemethod.getSource().getFile().getName().replace(".", "_");
				advicemethod.setName(newName);
			}
			if(codeStr.contains("object$rac")){
				String originalCall = "object$rac."+advicemethod.getCallSignature();
				adviceCode.append(originalCall).append(";\n");
				if(isReviseMeth || isAppendLikeMeth){
					advicemethod.setName(oldName);
				}
			}
			else{
				String originalCall = advicemethod.getParentClass().getFullyQualifiedName().replace("$", ".")+"."+
						advicemethod.getCallSignature();
				adviceCode.append(originalCall).append(";\n");
				if(isReviseMeth  || isAppendLikeMeth){
					advicemethod.setName(oldName);
				}
			}
			if(isAroundAdvice || isReviseMeth){
				if((advicemethod.getReturnType()!= null) && !advicemethod.getReturnType().getFullyQualifiedName().equals("void")){
					String pjpVarNAme = getProceedingJPVarDeclName(this.methodDecl.parameters());
					adviceCode.append("    ");
					adviceCode.append("else\n");
					adviceCode.append("    ");
					adviceCode.append(" return ");
					if(!advicemethod.getReturnType().getFullyQualifiedName().equals("java.lang.Object")){
//						adviceCode.append("(").append(advicemethod.getReturnType().getFullyQualifiedName()).append(")");
						adviceCode.append("(").append(this.methodDecl.returnType().toString()).append(")"); // this gets the erasure processed!!
					}
					adviceCode.append("$.proceed();\n".replace("$", pjpVarNAme));
				}
			}
		}
		adviceCode.append("}\n");
	}

	private String extractPCValue(String xcsPCTmp) {
		if(!(xcsPCTmp.contains("("))){
			for (Iterator iterator = AspectUtil.getInstance().getFieldDecls().iterator(); iterator.hasNext();) {
				JavaField javaField = (JavaField) iterator.next();
				if(xcsPCTmp.contains(".")){
					String field = javaField.getParentClass().getFullyQualifiedName().replace("$", ".")+"."+javaField.getName();
					if(field.equals(xcsPCTmp)){
						xcsPCTmp = javaField.getInitializationExpression().replace("\"", "");
					}
				}
				else{
					if(javaField.getParentClass().getFullyQualifiedName().replace("$", ".").
							equals(methodDecl.getMethod().owner().getJavaName().replace("$", "."))){
						if(javaField.getName().equals(xcsPCTmp)){
							xcsPCTmp = javaField.getInitializationExpression().replace("\"", "");
						}
					}
				}
			}
		}
		return xcsPCTmp;
	}

	private String processingXCSPointcut(String xcsParentClassName,
			String xcsParentClassFullyQualifiedName, Annotation adviceAnno,
			String xcsPCTmp, boolean flexibleXCS) {
		
		Pattern ownerTypePattern = Pattern.compile("\\b"+Pattern.quote(xcsParentClassFullyQualifiedName)+"\\b");
		Matcher ownerTypeMatcher = ownerTypePattern.matcher(xcsPCTmp);
		if(ownerTypeMatcher.find()){
			xcsPCTmp = xcsPCTmp.replaceAll("\\b"+Pattern.quote(xcsParentClassFullyQualifiedName)+"\\b", xcsParentClassName);
		}
		
		ownerTypePattern = Pattern.compile("\\b"+Pattern.quote(xcsParentClassName)+"\\b");
		ownerTypeMatcher = ownerTypePattern.matcher(xcsPCTmp);
		if(ownerTypeMatcher.find()){
			xcsPCTmp = xcsPCTmp.replaceAll("\\b"+Pattern.quote(xcsParentClassName)+"\\b", xcsParentClassFullyQualifiedName);
		}
		
		// handling execution pointcut
//		Pattern execPattern = Pattern.compile("execution(\\s)*\\((.)*\\)");
//		Matcher execMatcher = execPattern.matcher(xcsPCTmp);
//		if(execMatcher.find()){
//			if(Main.aRacOptions.clientAwareChecking() || Main.aRacOptions.noExecutionSiteInstrumentation()){
//				JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), adviceAnno.getLineNumber()), 
//						JmlMessages.INVALID_POINTCUT_EXEC, this.methodDecl.ident());
//				System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
//			}
//		}
		// handling call pointcut
//		Pattern callPattern = Pattern.compile("call(\\s)*\\((.)*\\)");
//		Matcher callMatcher = callPattern.matcher(xcsPCTmp);
//		if(callMatcher.find()){
//			if(!Main.aRacOptions.clientAwareChecking() && !Main.aRacOptions.callSiteInstrumentation()){
//				JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), adviceAnno.getLineNumber()), 
//						JmlMessages.INVALID_POINTCUT_CALL, this.methodDecl.ident());
//				System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
//			}
//		}
		
		if(!flexibleXCS){
			Pattern thisPattern = Pattern.compile("this(\\s)*\\((\\s)*");
			Matcher thisMatcher = thisPattern.matcher(xcsPCTmp);
			if(thisMatcher.find()){
				JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), adviceAnno.getLineNumber()), 
						JmlMessages.INVALID_POINTCUT_THIS, this.methodDecl.ident());
				System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
			}

			// handling target pointcut/designator
			Pattern targetPattern = Pattern.compile("target(\\s)*\\((\\s)*");
			Matcher targetMatcher = targetPattern.matcher(xcsPCTmp);
			if(targetMatcher.find()){
				JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), adviceAnno.getLineNumber()), 
						JmlMessages.INVALID_POINTCUT_TARGET, this.methodDecl.ident());
				System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
			}
		}
		// handling this pointcut/designator
	
		return xcsPCTmp;
	}

	/**
	 * @param code
	 * @param parameters
	 * @param classQualifiedName
	 * @param methodQualifiedName
	 */
	private String buildNPAndXPAroundAdviceHeader(JFormalParameter[] parameters, String classQualifiedName,
			String methodQualifiedName, String instrumentationType, long visibility, String methodQualifiedNameJP, 
			boolean isMethodCrosscutSpecChecking, String xcsPointcut, boolean isXCSPointcutStatic, boolean flexibleXCS) {
		StringBuffer code = new StringBuffer("");
		String packageQualifiedName = this.methodDecl.getMethod().owner().packageName().replace("/", ".");
//		JavaMethod jm = AspectUtil.getInstance().getCorrespondingJavaMethodThroughJMLMethod(this.methodDecl.getMethod().owner().getJavaName(), this.methodDecl);
		String methodReturnType = "";
//		if((jm != null) && (!(this.methodDecl.isConstructor()))){
//			if(jm.getReturnType().toString().equals(this.methodDecl.returnType().toString())){
//				methodReturnType = this.methodDecl.returnType().toString();
//			}
//			else{
//				methodReturnType = jm.getReturnType().toString();
//			}
//		}
//		else{
//			methodReturnType = this.methodDecl.returnType().toString();
//		}
		methodReturnType = this.methodDecl.returnType().toString();

		code.append("\n");
		if (this.javadoc != null) {
			if (instrumentationType.equals("clientAwareChecking")) {
				if (visibility == ACC_PUBLIC) {
					code.append(this.javadoc.replace("postcondition", "public postcondition"));
				} else if (visibility == ACC_PROTECTED) {
					code.append(this.javadoc.replace("postcondition", "protected postcondition"));
				} else if (visibility == 0L) { //default
					code.append(this.javadoc.replace("postcondition", "default postcondition"));
				} else if (visibility == ACC_PRIVATE) {
					code.append(this.javadoc.replace("postcondition", "private postcondition"));
				}
			}
			else{
				code.append(this.javadoc);
			}
		} else {
			code.append("/** Generated by JML to check assertions. */");
		}
		code.append("\n");
		
		if(isMethodCrosscutSpecChecking){
			if(isXCSPointcutStatic){
				code.append("java.lang.Object").append(" ").append("around (");
			}
			else{
				code.append("java.lang.Object").append(" ").append("around (final ").append(classQualifiedName).append(" ").append("object$rac");
			}
			code.append(generateAdviceMethodParameters(parameters, isXCSPointcutStatic, isMethodCrosscutSpecChecking, 
					AspectUtil.getInstance().isAspectAdvice(this.methodDecl)).toString());
			code.append(")");
//            code.append(this.generateMethodThrowsList(exceptionsInSignalsClauses));
			code.append(" :");
			code.append("\n");
			if(instrumentationType.equals("clientAwareChecking")){
				String call = isXCSPointcutStatic? "("+xcsPointcut+")" : "("+xcsPointcut+") && target("+classQualifiedName+")";
				AspectUtil.getInstance().appendPreconditionPointcut(call);
				code.append("   (").append(xcsPointcut).append(")");
				if(visibility == ACC_PROTECTED){
					code.append(" && ");
					code.append("\n");
					code.append("   (within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*) || within("+classQualifiedName+"+))");
				}
				else if(visibility == 0L){//package
					code.append(" && ");
					code.append("\n");
					code.append("   within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*)");
				}
				else if(visibility == ACC_PRIVATE){
					code.append(" && ");
					code.append("\n");
					code.append("   within("+classQualifiedName+")");
				}
				code.append(" && ");
				code.append("\n");
				code.append("   !within(*..AspectJMLRac_*) && !within(AspectJMLRac_*)");
			}
			else if(instrumentationType.equals("callSite")){
				String call = isXCSPointcutStatic? "("+xcsPointcut+")" : "("+xcsPointcut+") && target("+classQualifiedName+")";
				AspectUtil.getInstance().appendPreconditionPointcut(call);
				code.append("   (").append(xcsPointcut).append(")");
				code.append(" && ");
				code.append("\n");
				code.append("   !within(*..AspectJMLRac_*) && !within(AspectJMLRac_*)");
			}
			else{
				String execution = isXCSPointcutStatic? "("+xcsPointcut+")" : "("+xcsPointcut+") && target("+classQualifiedName+")";
				AspectUtil.getInstance().appendPreconditionPointcut(execution);
				code.append("   (").append(xcsPointcut).append(")");	
			}
			if(!isXCSPointcutStatic){
				code.append(" && ");
				code.append("\n");
				code.append("   target(object$rac)");
			}
			else{
				if(!flexibleXCS){
					code.append(" && ");
					code.append("\n");
					code.append("   if(JMLChecker.checkJP(thisJoinPoint.getSignature().getDeclaringType(), \"$\"))".replace("$", classQualifiedName));
				}
			}
			
			// appending pointcut for utilprecondition checkingXCS --- [[[hemr]]]
			String pc = "";
			if(instrumentationType.equals("clientAwareChecking")){
				pc = isXCSPointcutStatic? "("+xcsPointcut+")" : "("+xcsPointcut+") && target("+classQualifiedName+")";
			}
			else if(instrumentationType.equals("callSite")){
				pc = isXCSPointcutStatic? "("+xcsPointcut+")" : "("+xcsPointcut+") && target("+classQualifiedName+")";
			}
			else{
				pc = isXCSPointcutStatic? "("+xcsPointcut+")" : "("+xcsPointcut+") && target("+classQualifiedName+")";
			}
			AspectUtil.getInstance().appendPreconditionPointcut(pc);
			
		} // end of isMethodCrosscuttingSpecChecking
		else{
			String adviceexecution = "";
			String ifPC = "";
			if(AspectUtil.getInstance().isAspectAdvice(this.methodDecl)){
				adviceexecution = " || (adviceexecution(";
				String methodNameWithParameterTypes = methodDecl.ident()+AspectUtil.generateMethodParameters(methodDecl.parameters(), false).toString();
				ifPC = " && if(JMLChecker.getThisJoinPointPartialMethSig(thisJoinPoint.getSignature().toLongString(), " +
						"thisJoinPoint.getSignature().getDeclaringTypeName()).endsWith(\""+methodNameWithParameterTypes+"\"))))";
				
			}
			if(this.isStatic){
				code.append(methodReturnType).append(" ").append("around (");	
				code.append(generateAdviceMethodParameters(parameters, this.isStatic, isMethodCrosscutSpecChecking, 
						AspectUtil.getInstance().isAspectAdvice(this.methodDecl)).toString());
				code.append(")");
//	            code.append(this.generateMethodThrowsList(this.methodDecl.getExceptions()));
			}
			else{
				if(instrumentationType.equals("callSite") || instrumentationType.equals("clientAwareChecking")){
					if(this.methodDecl.isConstructor()){
						code.append(classQualifiedName).append(" ").append("around (");
						code.append(generateAdviceMethodParameters(parameters, true, isMethodCrosscutSpecChecking, 
								AspectUtil.getInstance().isAspectAdvice(this.methodDecl)).toString());
						code.append(")");
//						code.append(this.generateMethodThrowsList(this.methodDecl.getExceptions()));
					}
					else{
						code.append(methodReturnType).append(" ").append("around (final ").append(classQualifiedName).append(" ").append("object$rac");	
						code.append(generateAdviceMethodParameters(parameters, this.isStatic, isMethodCrosscutSpecChecking, 
								AspectUtil.getInstance().isAspectAdvice(this.methodDecl)).toString());
						code.append(")");
//						code.append(this.generateMethodThrowsList(this.methodDecl.getExceptions()));
					}
				}
				else{
					code.append(methodReturnType).append(" ").append("around (final ").append(classQualifiedName).append(" ").append("object$rac");
					code.append(generateAdviceMethodParameters(parameters, this.isStatic, isMethodCrosscutSpecChecking, 
							AspectUtil.getInstance().isAspectAdvice(this.methodDecl)).toString());
					code.append(")");
//					code.append(this.generateMethodThrowsList(this.methodDecl.getExceptions()));
				}	
			}
			code.append(" :");
			code.append("\n");
			if(this.methodDecl.isConstructor()){
				if(instrumentationType.equals("clientAwareChecking")){
					code.append("   call(").append(methodQualifiedName).append(")");
					if(visibility == ACC_PROTECTED){
						code.append(" && ");
						code.append("\n");
						code.append("   (within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*) || within("+classQualifiedName+"+))");
					}
					else if(visibility == 0L){//package
						code.append(" && ");
						code.append("\n");
						code.append("   within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*)");
					}
					else if(visibility == ACC_PRIVATE){
						code.append(" && ");
						code.append("\n");
						code.append("   within("+classQualifiedName+")");
					}
					code.append(" && ");
					code.append("\n");
					code.append("   !within(*..AspectJMLRac_*) && !within(AspectJMLRac_*)");
				}
				else if(instrumentationType.equals("callSite")){
					code.append("   call(").append(methodQualifiedName).append(")");
					code.append(" && ");
					code.append("\n");
					code.append("   !within(*..AspectJMLRac_*) && !within(AspectJMLRac_*)");
				}
				else{
					code.append("   execution(").append(methodQualifiedName).append(")");
				}
			}
			else if(this.isStatic){
				if(instrumentationType.equals("clientAwareChecking")){
					code.append("   call(").append("static").append(" ").append(methodReturnType).append(" ").append(methodQualifiedName).append("(").append(AspectUtil.generateMethodParametersForAdvicePointcut(parameters).toString()).append(")").append(")");
					if(visibility == ACC_PROTECTED){
						code.append(" && ");
						code.append("\n");
						code.append("   (within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*) || within("+classQualifiedName+"+))");
					}
					else if(visibility == 0L){//package
						code.append(" && ");
						code.append("\n");
						code.append("   within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*)");
					}
					else if(visibility == ACC_PRIVATE){
						code.append(" && ");
						code.append("\n");
						code.append("   within("+classQualifiedName+")");
					}
					code.append(" && ");
					code.append("\n");
					code.append("   !within(*..AspectJMLRac_*) && !within(AspectJMLRac_*)");
				}
				else if(instrumentationType.equals("callSite")){
					code.append("   call(").append("static").append(" ").append(methodReturnType).append(" ").append(methodQualifiedName).append("(").append(AspectUtil.generateMethodParametersForAdvicePointcut(parameters).toString()).append(")").append(")");
					code.append(" && ");
					code.append("\n");
					code.append("   !within(*..AspectJMLRac_*) && !within(AspectJMLRac_*)");
				}
				else{
					code.append("   execution(").append("static").append(" ").append(methodReturnType).append(" ").append(methodQualifiedName).append("(").append(AspectUtil.generateMethodParametersForAdvicePointcut(parameters).toString()).append(")").append(")");	
				}
			}
			else{
				if(instrumentationType.equals("clientAwareChecking")){
					String call = "(call("+methodReturnType+" "+methodQualifiedNameJP+"("+AspectUtil.generateMethodParametersForAdvicePointcut(parameters).toString()+"))"+adviceexecution+")";
					code.append("   (call(").append(methodReturnType).append(" ").append(methodQualifiedName).append("(").append(AspectUtil.generateMethodParametersForAdvicePointcut(parameters).toString()).
					append(")").append(")").append(adviceexecution).append(")").append(ifPC);
					AspectUtil.getInstance().appendPreconditionPointcut(call);
					if(visibility == ACC_PROTECTED){
						code.append(" && ");
						code.append("\n");
						code.append("   (within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*) || within("+classQualifiedName+"+))");
					}
					else if(visibility == 0L){//package
						code.append(" && ");
						code.append("\n");
						code.append("   within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*)");
					}
					else if(visibility == ACC_PRIVATE){
						code.append(" && ");
						code.append("\n");
						code.append("   within("+classQualifiedName+")");
					}
					code.append(" && ");
					code.append("\n");
					code.append("   !within(*..AspectJMLRac_*) && !within(AspectJMLRac_*)");
				}
				else if(instrumentationType.equals("callSite")){
					String call = "(call("+methodReturnType+" "+methodQualifiedNameJP+"("+AspectUtil.generateMethodParametersForAdvicePointcut(this.parameters).toString()+"))"+adviceexecution+")";
					AspectUtil.getInstance().appendPreconditionPointcut(call);
					code.append("   (call(").append(methodReturnType).append(" ").append(methodQualifiedName).append("(").append(AspectUtil.generateMethodParametersForAdvicePointcut(parameters).toString()).append(")").
					append(")").append(adviceexecution).append(")").append(ifPC);
					code.append(" && ");
					code.append("\n");
					code.append("   !within(*..AspectJMLRac_*) && !within(AspectJMLRac_*)");
				}
				else{
					String execution = "(execution("+methodReturnType+" "+methodQualifiedNameJP+"("+AspectUtil.generateMethodParametersForAdvicePointcut(this.parameters).toString()+"))"+adviceexecution+")";
					AspectUtil.getInstance().appendPreconditionPointcut(execution);
					code.append("   (execution(").append(methodReturnType).append(" ").append(methodQualifiedName).append("(").append(AspectUtil.generateMethodParametersForAdvicePointcut(parameters).toString()).append(")").
					append(")").append(adviceexecution).append(")").append(ifPC);	
				}
			}
			if(!this.isStatic){
				if(instrumentationType.equals("callSite") || instrumentationType.equals("clientAwareChecking")){
					if(!this.methodDecl.isConstructor()){
						code.append("\n");
						code.append("   && ");
						code.append("target(object$rac)");
					}
				}
				else{
					code.append("\n");
					code.append("   && ");
					code.append("this(object$rac)");	
				}
			}
		}
		
		if(!isMethodCrosscutSpecChecking){
			this.generateMethodArgsForTheAdvice(parameters, code, AspectUtil.getInstance().isAspectAdvice(this.methodDecl));
		}
		String result = code.toString();
		return result;
	}

	/**
	 * @param code
	 * @param parameters
	 * @param classQualifiedName
	 * @param methodQualifiedName
	 */
	private String buildNPAfterReturningAdviceHeader(JFormalParameter[] parameters, String classQualifiedName,
			String methodQualifiedName, String instrumentationType, long visibility, boolean isMethodCrosscutSpecChecking, 
			String xcsPointcut, boolean isXCSPointcutStatic, boolean flexibleXCS) {
		StringBuffer code = new StringBuffer("");
		String packageQualifiedName = this.methodDecl.getMethod().owner().packageName().replace("/", ".");
//		JavaMethod jm = AspectUtil.getInstance().getCorrespondingJavaMethodThroughJMLMethod(this.methodDecl.getMethod().owner().getJavaName(), this.methodDecl);
		String methodReturnType = "";
//		if((jm != null) && (!(this.methodDecl.isConstructor()))){
//			if(jm.getReturnType().toString().equals(this.methodDecl.returnType().toString())){
//				methodReturnType = this.methodDecl.returnType().toString();
//			}
//			else{
//				methodReturnType = jm.getReturnType().toString();
//			}
//		}
//		else{
//			methodReturnType = this.methodDecl.returnType().toString();
//		}
		methodReturnType = this.methodDecl.returnType().toString();
        
		code.append("\n");
		if (this.javadoc != null) {
			if (instrumentationType.equals("clientAwareChecking")) {
				if (visibility == ACC_PUBLIC) {
					code.append(this.javadoc.replace("postcondition", "public postcondition"));
				} else if (visibility == ACC_PROTECTED) {
					code.append(this.javadoc.replace("postcondition", "protected postcondition"));
				} else if (visibility == 0L) { //default
					code.append(this.javadoc.replace("postcondition", "default postcondition"));
				} else if (visibility == ACC_PRIVATE) {
					code.append(this.javadoc.replace("postcondition", "private postcondition"));
				}
			}
			else{
				code.append(this.javadoc);
			}
		} else {
			code.append("/** Generated by JML to check assertions. */");
		}
		code.append("\n");
		
		if(isMethodCrosscutSpecChecking){
			if(isXCSPointcutStatic){
				code.append("after (");
			}
			else{
				code.append("after (final ").append(classQualifiedName).append(" ").append("object$rac");
			}
			code.append(this.generateAdviceMethodParameters(parameters, isXCSPointcutStatic, isMethodCrosscutSpecChecking, 
					AspectUtil.getInstance().isAspectAdvice(this.methodDecl)).toString());
			code.append(")").append(" returning (final ").append("java.lang.Object").append(" rac$result) :");
			if(instrumentationType.equals("clientAwareChecking")){
				code.append("   (").append(xcsPointcut).append(")");
				if(visibility == ACC_PROTECTED){
					code.append(" && ");
					code.append("\n");
					code.append("   (within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*) || within("+classQualifiedName+"+))");
				}
				else if(visibility == 0L){//package
					code.append(" && ");
					code.append("\n");
					code.append("   within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*)");
				}
				else if(visibility == ACC_PRIVATE){
					code.append(" && ");
					code.append("\n");
					code.append("   within("+classQualifiedName+")");
				}
				code.append(" && ");
				code.append("\n");
				code.append("   !within(*..AspectJMLRac_*) && !within(AspectJMLRac_*)");
			}
			else if(instrumentationType.equals("callSite")){
				code.append("   (").append(xcsPointcut).append(")");
				code.append(" && ");
				code.append("\n");
				code.append("   !within(*..AspectJMLRac_*) && !within(AspectJMLRac_*)");
			}
			else{
				code.append("   (").append(xcsPointcut).append(")");
			}
			if(!isXCSPointcutStatic){
				code.append(" && ");
				code.append("\n");
				code.append("   target(object$rac)");
			}
			else{
				if(!flexibleXCS){
					code.append(" && ");
					code.append("\n");
					code.append("   if(JMLChecker.checkJP(thisJoinPoint.getSignature().getDeclaringType(), \"$\"))".replace("$", classQualifiedName));
				}
			}
		} // end crosscuting spec checking
		else{
			String adviceexecution = "";
			String ifPC = "";
			if(AspectUtil.getInstance().isAspectAdvice(this.methodDecl)){
				adviceexecution = " || (adviceexecution(";
				String methodNameWithParameterTypes = methodDecl.ident()+AspectUtil.generateMethodParameters(methodDecl.parameters(), false).toString();
				ifPC = " && if(JMLChecker.getThisJoinPointPartialMethSig(thisJoinPoint.getSignature().toLongString(), " +
						"thisJoinPoint.getSignature().getDeclaringTypeName()).endsWith(\""+methodNameWithParameterTypes+"\"))))";
				
			}
			if(this.isStatic){
				code.append(" ").append("after (");	
				code.append(this.generateAdviceMethodParameters(parameters, this.isStatic, isMethodCrosscutSpecChecking, 
						AspectUtil.getInstance().isAspectAdvice(this.methodDecl)).toString());
			}
			else{
				if(instrumentationType.equals("callSite") || instrumentationType.equals("clientAwareChecking")){
					if(this.methodDecl.isConstructor()){
						code.append("after (");
						code.append(this.generateAdviceMethodParameters(parameters, true, isMethodCrosscutSpecChecking, 
								AspectUtil.getInstance().isAspectAdvice(this.methodDecl)).toString());
					}
					else{
						code.append("after (final ").append(classQualifiedName).append(" ").append("object$rac");
						code.append(this.generateAdviceMethodParameters(parameters, this.isStatic, isMethodCrosscutSpecChecking, 
								AspectUtil.getInstance().isAspectAdvice(this.methodDecl)).toString());
					}
				}
				else{
					code.append("after (final ").append(classQualifiedName).append(" ").append("object$rac");
					code.append(this.generateAdviceMethodParameters(parameters, this.isStatic, isMethodCrosscutSpecChecking, 
							AspectUtil.getInstance().isAspectAdvice(this.methodDecl)).toString());
				}	
			}
			if(methodReturnType.equals("void")){
				if(this.methodDecl.isConstructor() && (instrumentationType.equals("callSite") || instrumentationType.equals("clientAwareChecking"))){
					code.append(")").append(" returning (final "+classQualifiedName+" object$rac) : ");
				}
				else{
					code.append(")").append(" returning () :");
				}
			}
			else{
				code.append(")").append(" returning (final "+this.methodDecl.returnType()+" rac$result) :");
			}
			code.append("\n");
			if(this.methodDecl.isConstructor()){
				if(instrumentationType.equals("clientAwareChecking")){
					code.append("   call(").append(methodQualifiedName).append(")");
					if(visibility == ACC_PROTECTED){
						code.append(" && ");
						code.append("\n");
						code.append("   (within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*) || within("+classQualifiedName+"+))");
					}
					else if(visibility == 0L){//package
						code.append(" && ");
						code.append("\n");
						code.append("   within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*)");
					}
					else if(visibility == ACC_PRIVATE){
						code.append(" && ");
						code.append("\n");
						code.append("   within("+classQualifiedName+")");
					}
					code.append(" && ");
					code.append("\n");
					code.append("   !within(*..AspectJMLRac_*) && !within(AspectJMLRac_*)");
				}
				else if(instrumentationType.equals("callSite")){
					code.append("   call(").append(methodQualifiedName).append(")");
					code.append(" && ");
					code.append("\n");
					code.append("   !within(*..AspectJMLRac_*) && !within(AspectJMLRac_*)");
				}
				else{
					code.append("   execution(").append(methodQualifiedName).append(")");
				}
			}
			else if(this.isStatic){
				if(instrumentationType.equals("clientAwareChecking")){
					code.append("   call(").append("static").append(" ").append(methodReturnType).append(" ").append(methodQualifiedName).append("(").append(AspectUtil.generateMethodParametersForAdvicePointcut(parameters).toString()).append(")").append(")");
					if(visibility == ACC_PROTECTED){
						code.append(" && ");
						code.append("\n");
						code.append("   (within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*) || within("+classQualifiedName+"+))");
					}
					else if(visibility == 0L){//package
						code.append(" && ");
						code.append("\n");
						code.append("   within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*)");
					}
					else if(visibility == ACC_PRIVATE){
						code.append(" && ");
						code.append("\n");
						code.append("   within("+classQualifiedName+")");
					}
					code.append(" && ");
					code.append("\n");
					code.append("   !within(*..AspectJMLRac_*) && !within(AspectJMLRac_*)");
				}
				else if(instrumentationType.equals("callSite")){
					code.append("   call(").append("static").append(" ").append(methodReturnType).append(" ").append(methodQualifiedName).append("(").append(AspectUtil.generateMethodParametersForAdvicePointcut(parameters).toString()).append(")").append(")");
					code.append(" && ");
					code.append("\n");
					code.append("   !within(*..AspectJMLRac_*) && !within(AspectJMLRac_*)");
				}
				else{
					code.append("   execution(").append("static").append(" ").append(methodReturnType).append(" ").append(methodQualifiedName).append("(").append(AspectUtil.generateMethodParametersForAdvicePointcut(parameters).toString()).append(")").append(")");	
				}
			}
			else{
				if(instrumentationType.equals("clientAwareChecking")){
					code.append("   (call(").append(methodReturnType).append(" ").append(methodQualifiedName).append("(").append(AspectUtil.generateMethodParametersForAdvicePointcut(parameters).toString()).
					append(")").append(")").append(adviceexecution).append(")").append(ifPC);
					if(visibility == ACC_PROTECTED){
						code.append(" && ");
						code.append("\n");
						code.append("   (within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*) || within("+classQualifiedName+"+))");
					}
					else if(visibility == 0L){//package
						code.append(" && ");
						code.append("\n");
						code.append("   within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*)");
					}
					else if(visibility == ACC_PRIVATE){
						code.append(" && ");
						code.append("\n");
						code.append("   within("+classQualifiedName+")");
					}
					code.append(" && ");
					code.append("\n");
					code.append("   !within(*..AspectJMLRac_*) && !within(AspectJMLRac_*)");
				}
				else if(instrumentationType.equals("callSite")){
					code.append("   (call(").append(methodReturnType).append(" ").append(methodQualifiedName).append("(").append(AspectUtil.generateMethodParametersForAdvicePointcut(parameters).toString()).
					append(")").append(")").append(adviceexecution).append(")").append(ifPC);
					code.append(" && ");
					code.append("\n");
					code.append("   !within(*..AspectJMLRac_*) && !within(AspectJMLRac_*)");
				}
				else{
					code.append("   (execution(").append(methodReturnType).append(" ").append(methodQualifiedName).append("(").append(AspectUtil.generateMethodParametersForAdvicePointcut(parameters).toString()).
					append(")").append(")").append(adviceexecution).append(")").append(ifPC);
				}
			}
			if(!this.isStatic){
				if(instrumentationType.equals("callSite") || instrumentationType.equals("clientAwareChecking")){
					if(!this.methodDecl.isConstructor()){
						code.append(" && ");
						code.append("target(object$rac)");
					}
				}
				else{
					code.append(" && ");
					code.append("this(object$rac)");
				}
			}
		}
		
		if(!isMethodCrosscutSpecChecking){
			this.generateMethodArgsForTheAdvice(parameters, code, AspectUtil.getInstance().isAspectAdvice(this.methodDecl));
		}
		String result = code.toString();
		return result;
	}
	
	private String buildBeforeAdviceHeaderForCC(String classQualifiedName, String xcsPointcut, boolean isAdviceMethodStatic, 
			boolean isAppendLikeMeth, boolean isAppendLikeMethCallPC, boolean adviceinterception, boolean reentering, boolean selfreentering) {
		StringBuffer code = new StringBuffer("");
		JavaMethod jm = AspectUtil.getInstance().getCorrespondingJavaMethodThroughJMLMethod(this.methodDecl.getMethod().owner().getJavaName(), this.methodDecl);
		
		code.append("\n");
		this.javadoc = "/** Generated by AspectJML to modularize a crosscutting structure " +
				"for\n" + 
				" * the Java members intercepted by " + this.methodDecl.ident() + " AspectJML/Advice. */";
		code.append(this.javadoc);
		code.append("\n");
//		// generating code for @before advice in @AspectJ
		code.append("@org.aspectj.lang.annotation.Before");
		code.append("(");
		if(isAppendLikeMeth){
			if(isAppendLikeMethCallPC) {
				code.append("\"").append("call(#)".replace("#", AspectUtil.getInstance().getFullyJavaMethodSignature(jm, false, false)));
			}
			else {
				code.append("\"").append("execution(#)".replace("#", AspectUtil.getInstance().getFullyJavaMethodSignature(jm, false, false)));
			}
			if(!jm.isStatic()){
				code.append(" && ");
				code.append("target(object$rac)");
				this.generateMethodArgsForTheAdvice(parameters, code, AspectUtil.getInstance().isAspectAdvice(this.methodDecl));
			}
			else{
				this.generateMethodArgsForTheAdvice(parameters, code, AspectUtil.getInstance().isAspectAdvice(this.methodDecl));
			}
		}
		else {
			// checking if the pointcut predicate come 
			// from a constant as a string
			if(!isAdviceMethodStatic){
				code.append("\"(").append(xcsPointcut);
				code.append(")");
				code.append(" && ");
				code.append("target(object$rac)");
				if(this.isCallPC(xcsPointcut) && this.hasConstructorInterception(xcsPointcut)){
					JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), jm.getLineNumber()), 
							JmlMessages.INVALID_ADVICE_DECLARATION_PC_CALL_NEW, "before", this.methodDecl.ident());
					System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
				}
			}
			else{
				code.append("\"").append(xcsPointcut);
			}
			if(!this.methodDecl.getMethod().isPublic()){
				String packageQualifiedName = this.methodDecl.getMethod().owner().packageName().replace("/", ".");
				if(this.methodDecl.getMethod().isPrivate()){
					code.append(" && ");
					code.append("within("+classQualifiedName+")");
				}
				else if(this.methodDecl.getMethod().isProtected()){
					code.append(" && ");
					code.append("(within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*) || within("+classQualifiedName+"+))");
				}
				else{ // package
					code.append(" && ");
					code.append("within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*)");
				}
			}
		}
		// Excluing aspects due to call pointcuts
		if(!isAppendLikeMeth){
			code.append(" && !within(*..AspectJMLInnerCC_*)");
			code.append(" && !within(AspectJMLInnerCC_*)");
			// Excluding join points marked as "ExcludeAdvising" --- [[[hemr]]]
			code.append(" && !@annotation(ExcludeAdvising)");
			code.append(" && !@annotation(Sealed)");
			if(!adviceinterception){
				code.append(" && !@annotation(Advice)");
			}
			if(!reentering && !(this.isCflowPC(xcsPointcut) || this.isCflowBelowPC(xcsPointcut))){
				String cflowPC = xcsPointcut.replaceFirst("(&&(\\s)*args\\((.)+\\))", "");
				if(this.isExecPC(xcsPointcut) && this.isCallPC(xcsPointcut)) {
					cflowPC = cflowPC.replace("call", "execution");
				}
				if(this.isExecPC(xcsPointcut)){
					cflowPC += " && !execution(static * !("+classQualifiedName+"+)"+".*(..))";
				}
				if(this.isCallPC(xcsPointcut)){
					cflowPC += " && !call(static * !("+classQualifiedName+"+)"+".*(..))";
				}
				code.append(" && !cflowbelow(").append(cflowPC).append(")");
			}
			if(!selfreentering){
				String cflowPC = "execution(#)".replace("#", AspectUtil.getInstance().getFullyJavaMethodSignatureWithOwnerName(jm));
				code.append(" && !cflow(").append(cflowPC).append(")");
			}
		}
		code.append("\"");
		code.append(")");
		code.append("\n");
		code.append("public void ");
		if(AspectUtil.getInstance().isAspectJMLAppendLikeMethod(jm)){
			code.append(this.methodDecl.ident()+"$AppendBefore");
		}
		else{
			code.append(this.methodDecl.ident());
		}
		if(isAdviceMethodStatic){
			code.append(" (");
		}
		else{
			code.append(" (final ").append(classQualifiedName).append(" ").append("object$rac");
		}
		String adviceParam = this.generateAdviceMethodParametersForAtAspectJ(this.parameters, isAdviceMethodStatic, true, 
				AspectUtil.getInstance().isAspectAdvice(this.methodDecl)).toString();
		code.append(adviceParam);
		if(isAdviceMethodStatic){
			if(!hasThisJPSpecialParamDecl(this.parameters)){
				if(adviceParam.equals("")){
					code.append("org.aspectj.lang.JoinPoint thisJP");	
				}
				else{
					code.append(", org.aspectj.lang.JoinPoint thisJP");
				}
			}
		}
		else{
			if(!hasThisJPSpecialParamDecl(this.parameters)){
				code.append(", org.aspectj.lang.JoinPoint thisJP");
			}
		}
		code.append(")");
		code.append(this.throwStmtsForCC());
		String result = code.toString();
		return result;
	}

	private boolean isCallPC(String xcsPointcut) {
		boolean result = false;
		Pattern execCallPattern = Pattern.compile("call(\\s)*\\(");
		Matcher execCallMatcher = execCallPattern.matcher(xcsPointcut);
		if(execCallMatcher.find()){
			result = true;
		}
		return result;
	}
	
	private boolean isExecPC(String xcsPointcut) {
		boolean result = false;
		Pattern pcPattern = Pattern.compile("execution(\\s)*\\(");
		Matcher pcMatcher = pcPattern.matcher(xcsPointcut);
		if(pcMatcher.find()){
			result = true;
		}
		return result;
	}
	
	private boolean isSetPC(String xcsPointcut) {
		boolean result = false;
		Pattern pcPattern = Pattern.compile("set(\\s)*\\(");
		Matcher pcMatcher = pcPattern.matcher(xcsPointcut);
		if(pcMatcher.find()){
			result = true;
		}
		return result;
	}
	
	private boolean isGetPC(String xcsPointcut) {
		boolean result = false;
		Pattern pcPattern = Pattern.compile("get(\\s)*\\(");
		Matcher pcMatcher = pcPattern.matcher(xcsPointcut);
		if(pcMatcher.find()){
			result = true;
		}
		return result;
	}
	
	private boolean isCflowPC(String xcsPointcut) {
		boolean result = false;
		Pattern pcPattern = Pattern.compile("cflow(\\s)*\\(");
		Matcher pcMatcher = pcPattern.matcher(xcsPointcut);
		if(pcMatcher.find()){
			result = true;
		}
		return result;
	}
	
	private boolean isCflowBelowPC(String xcsPointcut) {
		boolean result = false;
		Pattern pcPattern = Pattern.compile("cflowbelow(\\s)*\\(");
		Matcher pcMatcher = pcPattern.matcher(xcsPointcut);
		if(pcMatcher.find()){
			result = true;
		}
		return result;
	}
	
	private boolean hasConstructorInterception(String xcsPointcut) {
		boolean result = false;
		Pattern execNewPattern = Pattern.compile("(\\b)new(\\b)");
		Matcher execNewMatcher = execNewPattern.matcher(xcsPointcut);
		if(execNewMatcher.find()){
			result = true;
		}
		return result;
	}
	
	private String buildAroundAdviceHeaderForCC(JFormalParameter[] parameters, String classQualifiedName,
			String xcsPointcut, boolean isAdviceMethodStatic, boolean isRevise, boolean isReviseCallPC, boolean adviceinterception, boolean reentering, boolean selfreentering) {
		StringBuffer code = new StringBuffer("");
		JavaMethod jm = AspectUtil.getInstance().getCorrespondingJavaMethodThroughJMLMethod(this.methodDecl.getMethod().owner().getJavaName(), this.methodDecl);
		String methodReturnType = "";
		if(isReviseCallPC && jm.isConstructor()) {
			JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), jm.getLineNumber()), 
					JmlMessages.INVALID_REVISE_AND_CALLSITE, this.methodDecl.ident());
			System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
		}
//		if((jm != null) && (!(this.methodDecl.isConstructor()))){
//			if(jm.getReturnType().toString().equals(this.methodDecl.returnType().toString())){
//				methodReturnType = this.methodDecl.returnType().toString();
//			}
//			else{
//				methodReturnType = jm.getReturnType().toString();
//			}
//		}
//		else{
//			methodReturnType = this.methodDecl.returnType().toString();
//		}
		methodReturnType = this.methodDecl.returnType().toString();

		code.append("\n");
		this.javadoc = "/** Generated by AspectJML to modularize a crosscutting structure " +
				"for\n" + 
				" * the Java members intercepted by " + this.methodDecl.ident() + " AspectJML/Advice. */";
		code.append(this.javadoc);
		code.append("\n");
//		// generating code for @around advice in @AspectJ
		code.append("@org.aspectj.lang.annotation.Around");
		code.append("(");
		if(isRevise){
			if(isReviseCallPC) {
				code.append("\"").append("call(#)".replace("#", AspectUtil.getInstance().getFullyJavaMethodSignature(jm, false, false)));
			}
			else {
				code.append("\"").append("execution(#)".replace("#", AspectUtil.getInstance().getFullyJavaMethodSignature(jm, false, false)));
			}
			if(!jm.isStatic()){
				code.append(" && ");
				code.append("target(object$rac)");
				this.generateMethodArgsForTheAdvice(parameters, code, AspectUtil.getInstance().isAspectAdvice(this.methodDecl));
			}
			else{
				this.generateMethodArgsForTheAdvice(parameters, code, AspectUtil.getInstance().isAspectAdvice(this.methodDecl));
			}
		}
		else{
			// checking if the pointcut predicate come 
			// from a constant as a string
			if(!isAdviceMethodStatic){
				code.append("\"(").append(xcsPointcut);
				code.append(")");
				code.append(" && ");
				code.append("target(object$rac)");
				if(this.isCallPC(xcsPointcut) && this.hasConstructorInterception(xcsPointcut)){
					JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), jm.getLineNumber()), 
							JmlMessages.INVALID_ADVICE_DECLARATION_PC_CALL_NEW, "around", this.methodDecl.ident());
					System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
				}
			}
			else{
				code.append("\"").append(xcsPointcut);
			}
			if(!this.methodDecl.getMethod().isPublic()){
				String packageQualifiedName = this.methodDecl.getMethod().owner().packageName().replace("/", ".");
				if(this.methodDecl.getMethod().isPrivate()){
					code.append(" && ");
					code.append("within("+classQualifiedName+")");
				}
				else if(this.methodDecl.getMethod().isProtected()){
					code.append(" && ");
					code.append("(within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*) || within("+classQualifiedName+"+))");
				}
				else{ // package
					code.append(" && ");
					code.append("within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*)");
				}
			}
		}
		if(!isRevise){
			// Excluing aspects due to call pointcuts
			code.append(" && !within(*..AspectJMLInnerCC_*)");
			code.append(" && !within(AspectJMLInnerCC_*)");
			// Excluding join points marked as "ExcludeAdvising" --- [[[hemr]]]
			code.append(" && !@annotation(ExcludeAdvising)");
			code.append(" && !@annotation(Sealed)");
			if(!adviceinterception){
				code.append(" && !@annotation(Advice)");
			}
			if(!reentering && !(this.isCflowPC(xcsPointcut) || this.isCflowBelowPC(xcsPointcut))){
				String cflowPC = xcsPointcut.replaceFirst("(&&(\\s)*args\\((.)+\\))", "");
				if(this.isExecPC(xcsPointcut) && this.isCallPC(xcsPointcut)) {
					cflowPC = cflowPC.replace("call", "execution");
				}
				if(this.isExecPC(xcsPointcut)){
					cflowPC += " && !execution(static * !("+classQualifiedName+"+)"+".*(..))";
				}
				if(this.isCallPC(xcsPointcut)){
					cflowPC += " && !call(static * !("+classQualifiedName+"+)"+".*(..))";
				}
				code.append(" && !cflowbelow(").append(cflowPC).append(")");
			}
			if(!selfreentering){
				String cflowPC = "execution(#)".replace("#", AspectUtil.getInstance().getFullyJavaMethodSignatureWithOwnerName(jm));
				code.append(" && !cflow(").append(cflowPC).append(")");
			}
		}
		code.append("\"");
		code.append(")");
		code.append("\n");
		code.append("public ");
		code.append(methodReturnType+" ");
		if(isRevise){
			if(jm.isConstructor()){
				code.append("new$Revise");
			}
			else{
				code.append(this.methodDecl.ident()+"$Revise");
			}
		}
		else{
			code.append(this.methodDecl.ident());
		}
		
		if(isRevise){
			if(jm.isStatic()){
				code.append(" (");
			}
			else{
				code.append(" (final ").append(classQualifiedName).append(" ").append("object$rac");
			}
		}
		else{
			if(isAdviceMethodStatic){
				code.append(" (");
			}
			else{
				code.append(" (final ").append(classQualifiedName).append(" ").append("object$rac");
			}
		}
		
		String adviceParam = this.generateAdviceMethodParametersForAtAspectJ(this.parameters, isAdviceMethodStatic, true, 
				AspectUtil.getInstance().isAspectAdvice(this.methodDecl)).toString();
		code.append(adviceParam);
		if(isAdviceMethodStatic){
			if(!hasProceedingJPSpecialParamDecl(this.parameters)){
				if(adviceParam.equals("")){
					code.append("org.aspectj.lang.ProceedingJoinPoint pjp");	
				}
				else{
					code.append(", org.aspectj.lang.ProceedingJoinPoint pjp");
				}
			}
		}
		else{
			if(!hasProceedingJPSpecialParamDecl(this.parameters)){
				code.append(", org.aspectj.lang.ProceedingJoinPoint pjp");
			}
		}
		code.append(")");
		code.append(this.throwStmtsForCC());
			
		String result = code.toString();
		return result;
	}
	
	private String buildAfterAdviceHeaderForCC(JFormalParameter[] parameters, String classQualifiedName,
			String methodQualifiedName, String xcsPointcut, boolean isAdviceMethodStatic, boolean isAppendLikeMeth, boolean isAppendLikeMethCallPC, boolean adviceinterception, boolean reentering, boolean selfreentering) {
		StringBuffer code = new StringBuffer("");
		JavaMethod jm = AspectUtil.getInstance().getCorrespondingJavaMethodThroughJMLMethod(this.methodDecl.getMethod().owner().getJavaName(), this.methodDecl);
		
		code.append("\n");
		this.javadoc = "/** Generated by AspectJML to modularize a crosscutting structure " +
				"for\n" + 
				" * the Java members intercepted by " + this.methodDecl.ident() + " AspectJML/Advice. */";
		code.append(this.javadoc);
		code.append("\n");
//		// generating code for @after advice in @AspectJ
		code.append("@org.aspectj.lang.annotation.After");
		code.append("(");
		if(isAppendLikeMeth){
			if(isAppendLikeMethCallPC) {
				code.append("\"").append("call(#)".replace("#", AspectUtil.getInstance().getFullyJavaMethodSignature(jm, false, false)));
			}
			else {
				code.append("\"").append("execution(#)".replace("#", AspectUtil.getInstance().getFullyJavaMethodSignature(jm, false, false)));
			}
			if(!jm.isStatic()){
				code.append(" && ");
				code.append("target(object$rac)");
				this.generateMethodArgsForTheAdvice(parameters, code, AspectUtil.getInstance().isAspectAdvice(this.methodDecl));
			}
			else{
				this.generateMethodArgsForTheAdvice(parameters, code, AspectUtil.getInstance().isAspectAdvice(this.methodDecl));
			}
		}
		else {
			// checking if the pointcut predicate come 
			// from a constant as a string
			if(!isAdviceMethodStatic){
				code.append("\"(").append(xcsPointcut);
				code.append(")");
				code.append(" && ");
				code.append("target(object$rac)");
				if(this.isCallPC(xcsPointcut) && this.hasConstructorInterception(xcsPointcut)){
					JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), jm.getLineNumber()), 
							JmlMessages.INVALID_ADVICE_DECLARATION_PC_CALL_NEW, "after", this.methodDecl.ident());
					System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
				}
			}
			else{
				code.append("\"").append(xcsPointcut);
			}
			if(!this.methodDecl.getMethod().isPublic()){
				String packageQualifiedName = this.methodDecl.getMethod().owner().packageName().replace("/", ".");
				if(this.methodDecl.getMethod().isPrivate()){
					code.append(" && ");
					code.append("within("+classQualifiedName+")");
				}
				else if(this.methodDecl.getMethod().isProtected()){
					code.append(" && ");
					code.append("(within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*) || within("+classQualifiedName+"+))");
				}
				else{ // package
					code.append(" && ");
					code.append("within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*)");
				}
			}
		}
		// Excluing aspects due to call pointcuts
		if(!isAppendLikeMeth){
			code.append(" && !within(*..AspectJMLInnerCC_*)");
			code.append(" && !within(AspectJMLInnerCC_*)");
			// Excluding join points marked as "ExcludeAdvising" --- [[[hemr]]]
			code.append(" && !@annotation(ExcludeAdvising)");
			code.append(" && !@annotation(Sealed)");
			if(!adviceinterception){
				code.append(" && !@annotation(Advice)");
			}
			if(!reentering && !(this.isCflowPC(xcsPointcut) || this.isCflowBelowPC(xcsPointcut))){
				String cflowPC = xcsPointcut.replaceFirst("(&&(\\s)*args\\((.)+\\))", "");
				if(this.isExecPC(xcsPointcut) && this.isCallPC(xcsPointcut)) {
					cflowPC = cflowPC.replace("call", "execution");
				}
				if(this.isExecPC(xcsPointcut)){
					cflowPC += " && !execution(static * !("+classQualifiedName+"+)"+".*(..))";
				}
				if(this.isCallPC(xcsPointcut)){
					cflowPC += " && !call(static * !("+classQualifiedName+"+)"+".*(..))";
				}
				code.append(" && !cflowbelow(").append(cflowPC).append(")");
			}
			if(!selfreentering){
				String cflowPC = "execution(#)".replace("#", AspectUtil.getInstance().getFullyJavaMethodSignatureWithOwnerName(jm));
				code.append(" && !cflow(").append(cflowPC).append(")");
			}
		}
		code.append("\"");
		code.append(")");
		code.append("\n");
		code.append("public void ");
		if(AspectUtil.getInstance().isAspectJMLAppendLikeMethod(jm)){
			code.append(this.methodDecl.ident()+"$AppendAfter");
		}
		else{
			code.append(this.methodDecl.ident());
		}
		
		if(isAdviceMethodStatic){
			code.append(" (");
		}
		else{
			code.append(" (final ").append(classQualifiedName).append(" ").append("object$rac");
		}
		String adviceParam = this.generateAdviceMethodParametersForAtAspectJ(this.parameters, isAdviceMethodStatic, true, 
				AspectUtil.getInstance().isAspectAdvice(this.methodDecl)).toString();
		code.append(adviceParam);
		if(isAdviceMethodStatic){
			if(!hasThisJPSpecialParamDecl(this.parameters)){
				if(adviceParam.equals("")){
					code.append("org.aspectj.lang.JoinPoint thisJP");	
				}
				else{
					code.append(", org.aspectj.lang.JoinPoint thisJP");
				}
			}
		}
		else{
			if(!hasThisJPSpecialParamDecl(this.parameters)){
				code.append(", org.aspectj.lang.JoinPoint thisJP");
			}
		}
		code.append(")");
		code.append(this.throwStmtsForCC());
		String result = code.toString();
		return result;
	}
	
	private String buildAfterReturningAdviceHeaderForCC(JFormalParameter[] parameters, String classQualifiedName, String xcsPointcut, 
			String xcsReturningOrThrowProperty, boolean isAdviceMethodStatic, boolean isAppendLikeMeth, boolean isAppendLikeMethCallPC, boolean adviceinterception, boolean reentering, boolean selfreentering) {
		StringBuffer code = new StringBuffer("");
		JavaMethod jm = AspectUtil.getInstance().getCorrespondingJavaMethodThroughJMLMethod(this.methodDecl.getMethod().owner().getJavaName(), this.methodDecl);
		
		code.append("\n");
		this.javadoc = "/** Generated by AspectJML to modularize a crosscutting structure " +
				"for\n" + 
				" * the Java members intercepted by " + this.methodDecl.ident() + " AspectJML/Advice. */";
		code.append(this.javadoc);
		code.append("\n");
//		// generating code for @after returning advice in @AspectJ
		code.append("@org.aspectj.lang.annotation.AfterReturning");
		code.append("(");
		if(isAppendLikeMeth){
			if(isAppendLikeMethCallPC) {
				code.append("\"").append("call(#)".replace("#", AspectUtil.getInstance().getFullyJavaMethodSignature(jm, false, false)));
			}
			else {
				code.append("\"").append("execution(#)".replace("#", AspectUtil.getInstance().getFullyJavaMethodSignature(jm, false, false)));
			}
			if(!jm.isStatic()){
				code.append(" && ");
				code.append("target(object$rac)");
				this.generateMethodArgsForTheAdvice(parameters, code, AspectUtil.getInstance().isAspectAdvice(this.methodDecl));
			}
			else{
				this.generateMethodArgsForTheAdvice(parameters, code, AspectUtil.getInstance().isAspectAdvice(this.methodDecl));
			}
		}
		else {
			// checking if the pointcut predicate come 
			// from a constant as a string
			if(!isAdviceMethodStatic){
				if(this.isCallPC(xcsPointcut) && this.hasConstructorInterception(xcsPointcut)){
					code.append("returning= \"object$rac\", ");
				}
				if(!xcsReturningOrThrowProperty.equals("")){
					if(this.isCallPC(xcsPointcut) && this.hasConstructorInterception(xcsPointcut)){
						//error
						JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), jm.getLineNumber()), 
								JmlMessages.INVALID_AFTER_RETURNING_ADVICE_DECLARATION_WITH_RETURNING_PROPERTY, "after-returning", this.methodDecl.ident());
						System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
					}
					code.append("pointcut=");
				}
				else{
					if(this.isCallPC(xcsPointcut) && this.hasConstructorInterception(xcsPointcut)){
						code.append("pointcut=");
					}
				}
				code.append("\"(").append(xcsPointcut);
				code.append(")");
				if(!(this.isCallPC(xcsPointcut) && this.hasConstructorInterception(xcsPointcut))){
					code.append(" && ");
					code.append("target(object$rac)");
				}
				
				if(!this.methodDecl.getMethod().isPublic()){
					String packageQualifiedName = this.methodDecl.getMethod().owner().packageName().replace("/", ".");
					if(this.methodDecl.getMethod().isPrivate()){
						code.append(" && ");
						code.append("within("+classQualifiedName+")");
					}
					else if(this.methodDecl.getMethod().isProtected()){
						code.append(" && ");
						code.append("(within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*) || within("+classQualifiedName+"+))");
					}
					else{ // package
						code.append(" && ");
						code.append("within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*)");
					}
				}
			}
			else{
				if(!xcsReturningOrThrowProperty.equals("")){
					code.append("pointcut=");
				}
				code.append("\"").append(xcsPointcut);
				if(!this.methodDecl.getMethod().isPublic()){
					String packageQualifiedName = this.methodDecl.getMethod().owner().packageName().replace("/", ".");
					if(this.methodDecl.getMethod().isPrivate()){
						code.append(" && ");
						code.append("within("+classQualifiedName+")");
					}
					else if(this.methodDecl.getMethod().isProtected()){
						code.append(" && ");
						code.append("(within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*) || within("+classQualifiedName+"+))");
					}
					else{ // package
						code.append(" && ");
						code.append("within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*)");
					}
				}
			}
		}
		// Excluing aspects due to call pointcuts
		if(!isAppendLikeMeth){
			code.append(" && !within(*..AspectJMLInnerCC_*)");
			code.append(" && !within(AspectJMLInnerCC_*)");
			// Excluding join points marked as "ExcludeAdvising" --- [[[hemr]]]
			code.append(" && !@annotation(ExcludeAdvising)");
			code.append(" && !@annotation(Sealed)");
			if(!adviceinterception){
				code.append(" && !@annotation(Advice)");
			}
			if(!reentering && !(this.isCflowPC(xcsPointcut) || this.isCflowBelowPC(xcsPointcut))){
				String cflowPC = xcsPointcut.replaceFirst("(&&(\\s)*args\\((.)+\\))", "");
				if(this.isExecPC(xcsPointcut) && this.isCallPC(xcsPointcut)) {
					cflowPC = cflowPC.replace("call", "execution");
				}
				if(this.isExecPC(xcsPointcut)){
					cflowPC += " && !execution(static * !("+classQualifiedName+"+)"+".*(..))";
				}
				if(this.isCallPC(xcsPointcut)){
					cflowPC += " && !call(static * !("+classQualifiedName+"+)"+".*(..))";
				}
				code.append(" && !cflowbelow(").append(cflowPC).append(")");
			}
			if(!selfreentering){
				String cflowPC = "execution(#)".replace("#", AspectUtil.getInstance().getFullyJavaMethodSignatureWithOwnerName(jm));
				code.append(" && !cflow(").append(cflowPC).append(")");
			}
		}
		code.append("\"");
		if(!xcsReturningOrThrowProperty.equals("")){
			code.append(", ");
			code.append(xcsReturningOrThrowProperty);
		}
		code.append(")");
		code.append("\n");
		code.append("public void ");
		if(AspectUtil.getInstance().isAspectJMLAppendLikeMethod(jm)){
			code.append(this.methodDecl.ident()+"$AppendAfterReturning");
		}
		else{
			code.append(this.methodDecl.ident());
		}
		if(isAdviceMethodStatic){
			code.append(" (");
		}
		else{
			code.append(" (final ").append(classQualifiedName).append(" ").append("object$rac");
		}
		String adviceParam = this.generateAdviceMethodParametersForAtAspectJ(this.parameters, isAdviceMethodStatic, true, 
				AspectUtil.getInstance().isAspectAdvice(this.methodDecl)).toString();
		code.append(adviceParam);
		if(isAdviceMethodStatic){
			if(!hasThisJPSpecialParamDecl(this.parameters)){
				if(adviceParam.equals("")){
					code.append("org.aspectj.lang.JoinPoint thisJP");	
				}
				else{
					code.append(", org.aspectj.lang.JoinPoint thisJP");
				}
			}
		}
		else{
			if(!hasThisJPSpecialParamDecl(this.parameters)){
				code.append(", org.aspectj.lang.JoinPoint thisJP");
			}
		}
		code.append(")");
		code.append(this.throwStmtsForCC());
		
//      old code for standard AspectJ after returning advice
//		code.append("\n");
//		this.javadoc = "/** Generated by AspectJML to modularize a crosscutting structure " +
//				"for\n" + 
//				" * the Java members intercepted by " + this.methodDecl.ident() + " AspectJML/Advice. */";
//		code.append(this.javadoc);
//		code.append("\n");
//
//		if(isXCSPointcutStatic){
//			code.append("after (");
//		}
//		else{
//			code.append("after (final ").append(classQualifiedName).append(" ").append("object$rac");
//		}
//		code.append(this.generateAdviceMethodParameters(parameters, isXCSPointcutStatic, true, 
//				AspectUtil.getInstance().isAspectAdvice(this.methodDecl)).toString());
//		code.append(")").append(" returning (final ").append("java.lang.Object").append(" rac$result) :\n");
//		code.append("   (").append(xcsPointcut).append(")");
//		if(!isXCSPointcutStatic){
//			code.append(" && ");
//			code.append("\n");
//			code.append("   target(object$rac)");	
//		}

		String result = code.toString();
		return result;
	}
	
	private String buildAfterThrowingAdviceHeaderForCC(JFormalParameter[] parameters, String classQualifiedName,
			String xcsPointcut, String xcsReturningOrThrowProperty, boolean isAdviceMethodStatic, boolean isAppendLikeMeth, boolean isAppendLikeMethCallPC, 
			boolean adviceinterception, boolean reentering, boolean selfreentering) {
		StringBuffer code = new StringBuffer("");
		JavaMethod jm = AspectUtil.getInstance().getCorrespondingJavaMethodThroughJMLMethod(this.methodDecl.getMethod().owner().getJavaName(), this.methodDecl);
		
		code.append("\n");
		this.javadoc = "/** Generated by AspectJML to modularize a crosscutting structure " +
				"for\n" + 
				" * the Java members intercepted by " + this.methodDecl.ident() + " AspectJML/Advice. */";
		code.append(this.javadoc);
		code.append("\n");
//		// generating code for @after throwing advice in @AspectJ
		code.append("@org.aspectj.lang.annotation.AfterThrowing");
		code.append("(");
		if(isAppendLikeMeth){
			if(isAppendLikeMethCallPC) {
				code.append("\"").append("call(#)".replace("#", AspectUtil.getInstance().getFullyJavaMethodSignature(jm, false, false)));
			}
			else {
				code.append("\"").append("execution(#)".replace("#", AspectUtil.getInstance().getFullyJavaMethodSignature(jm, false, false)));
			}
			if(!jm.isStatic()){
				code.append(" && ");
				code.append("target(object$rac)");
				this.generateMethodArgsForTheAdvice(parameters, code, AspectUtil.getInstance().isAspectAdvice(this.methodDecl));
			}
			else{
				this.generateMethodArgsForTheAdvice(parameters, code, AspectUtil.getInstance().isAspectAdvice(this.methodDecl));
			}
		}
		else {
			// checking if the pointcut predicate come 
			// from a constant as a string
			if(!isAdviceMethodStatic){
				if(!xcsReturningOrThrowProperty.equals("")){
					code.append("pointcut=");
				}
				code.append("\"(").append(xcsPointcut);
				code.append(")");
				code.append(" && ");
				code.append("target(object$rac)");
				if(this.isCallPC(xcsPointcut) && this.hasConstructorInterception(xcsPointcut)){
					JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), jm.getLineNumber()), 
							JmlMessages.INVALID_ADVICE_DECLARATION_PC_CALL_NEW, "after-throwing", this.methodDecl.ident());
					System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
				}
				if(!this.methodDecl.getMethod().isPublic()){
					String packageQualifiedName = this.methodDecl.getMethod().owner().packageName().replace("/", ".");
					if(this.methodDecl.getMethod().isPrivate()){
						code.append(" && ");
						code.append("within("+classQualifiedName+")");
					}
					else if(this.methodDecl.getMethod().isProtected()){
						code.append(" && ");
						code.append("(within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*) || within("+classQualifiedName+"+))");
					}
					else{ // package
						code.append(" && ");
						code.append("within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*)");
					}
				}
			}
			else{
				if(!xcsReturningOrThrowProperty.equals("")){
					code.append("pointcut=");
				}
				code.append("\"").append(xcsPointcut);
				if(!this.methodDecl.getMethod().isPublic()){
					String packageQualifiedName = this.methodDecl.getMethod().owner().packageName().replace("/", ".");
					if(this.methodDecl.getMethod().isPrivate()){
						code.append(" && ");
						code.append("within("+classQualifiedName+")");
					}
					else if(this.methodDecl.getMethod().isProtected()){
						code.append(" && ");
						code.append("(within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*) || within("+classQualifiedName+"+))");
					}
					else{ // package
						code.append(" && ");
						code.append("within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*)");
					}
				}
			}
		}
		// Excluing aspects due to call pointcuts
		if(!isAppendLikeMeth){
			code.append(" && !within(*..AspectJMLInnerCC_*)");
			code.append(" && !within(AspectJMLInnerCC_*)");
			// Excluding join points marked as "ExcludeAdvising" --- [[[hemr]]]
			code.append(" && !@annotation(ExcludeAdvising)");
			code.append(" && !@annotation(Sealed)");
			if(!adviceinterception){
				code.append(" && !@annotation(Advice)");
			}
			if(!reentering && !(this.isCflowPC(xcsPointcut) || this.isCflowBelowPC(xcsPointcut))){
				String cflowPC = xcsPointcut.replaceFirst("(&&(\\s)*args\\((.)+\\))", "");
				if(this.isExecPC(xcsPointcut) && this.isCallPC(xcsPointcut)) {
					cflowPC = cflowPC.replace("call", "execution");
				}
				if(this.isExecPC(xcsPointcut)){
					cflowPC += " && !execution(static * !("+classQualifiedName+"+)"+".*(..))";
				}
				if(this.isCallPC(xcsPointcut)){
					cflowPC += " && !call(static * !("+classQualifiedName+"+)"+".*(..))";
				}
				code.append(" && !cflowbelow(").append(cflowPC).append(")");
			}
			if(!selfreentering){
				String cflowPC = "execution(#)".replace("#", AspectUtil.getInstance().getFullyJavaMethodSignatureWithOwnerName(jm));
				code.append(" && !cflow(").append(cflowPC).append(")");
			}
		}
		code.append("\"");
		if(!xcsReturningOrThrowProperty.equals("")){
			code.append(", ");
			code.append(xcsReturningOrThrowProperty);
		}
		code.append(")");
		code.append("\n");
		code.append("public void ");
		if(AspectUtil.getInstance().isAspectJMLAppendLikeMethod(jm)){
			code.append(this.methodDecl.ident()+"$AppendAfterThrowing");
		}
		else{
			code.append(this.methodDecl.ident());
		}
		if(isAdviceMethodStatic){
			code.append(" (");
		}
		else{
			code.append(" (final ").append(classQualifiedName).append(" ").append("object$rac");
		}
		String adviceParam = this.generateAdviceMethodParametersForAtAspectJ(this.parameters, isAdviceMethodStatic, true, 
				AspectUtil.getInstance().isAspectAdvice(this.methodDecl)).toString();
		code.append(adviceParam);
		if(isAdviceMethodStatic){
			if(!hasThisJPSpecialParamDecl(this.parameters)){
				if(adviceParam.equals("")){
					code.append("org.aspectj.lang.JoinPoint thisJP");	
				}
				else{
					code.append(", org.aspectj.lang.JoinPoint thisJP");
				}
			}
		}
		else{
			if(!hasThisJPSpecialParamDecl(this.parameters)){
				code.append(", org.aspectj.lang.JoinPoint thisJP");
			}
		}
		code.append(")");
		code.append(this.throwStmtsForCC());
		
//      old code for standard AspectJ after throwing advice
//		code.append("\n");
//		this.javadoc = "/** Generated by AspectJML to modularize a crosscutting structure " +
//				"for\n" + 
//				" * the Java members intercepted by " + this.methodDecl.ident() + " AspectJML/Advice. */";
//		code.append(this.javadoc);
//		code.append(this.javadoc);
//		code.append("\n");
//
//		if(isXCSPointcutStatic){
//			code.append("after (");
//		}
//		else{
//			code.append("after (final ").append(classQualifiedName).append(" ").append("object$rac");
//		}
//		code.append(this.generateAdviceMethodParameters(parameters, isXCSPointcutStatic, true, 
//				AspectUtil.getInstance().isAspectAdvice(this.methodDecl)).toString());
//		code.append(")").append(" throwing (Throwable rac$e)");
//		//			code.append(this.generateMethodThrowsList(exceptionsInSignalsClauses));
//		code.append(" :");
//		code.append("\n");
//		code.append("   (").append(xcsPointcut).append(")");	
//		if(!isXCSPointcutStatic){
//			code.append(" && ");
//			code.append("\n");
//			code.append("   target(object$rac)");
//		}

		String result = code.toString();
		return result;
	}

	/**
	 * @param code
	 * @param parameters
	 * @param classQualifiedName
	 * @param methodQualifiedName
	 */
	private String buildXPAfterThrowingAdviceHeader(JFormalParameter[] parameters, String classQualifiedName,
			String methodQualifiedName, String instrumentationType, long visibility, boolean isMethodCrosscutSpecChecking, 
			String xcsPointcut, boolean isXCSPointcutStatic, boolean flexibleXCS) {
		StringBuffer code = new StringBuffer("");
		String packageQualifiedName = this.methodDecl.getMethod().owner().packageName().replace("/", ".");
//		JavaMethod jm = AspectUtil.getInstance().getCorrespondingJavaMethodThroughJMLMethod(this.methodDecl.getMethod().owner().getJavaName(), this.methodDecl);
		String methodReturnType = "";
//		if((jm != null) && (!(this.methodDecl.isConstructor()))){
//			if(jm.getReturnType() == null){
//				JmlRacGenerator.fail(TokenReference.build(this.methodDecl.getTokenReference().file(), jm.getLineNumber()), 
//						JmlMessages.INVALID_CONSTRUCTOR_RETURN_TYPE, this.methodDecl.ident(), 
//						classQualifiedName);
//				System.exit(0); // just because of the duplication of error messages ... [[[hemr]]]
//			}
//			if(jm.getReturnType().toString().equals(this.methodDecl.returnType().toString())){
//				methodReturnType = this.methodDecl.returnType().toString();
//			}
//			else{
//				methodReturnType = jm.getReturnType().toString();
//			}
//		}
//		else{
//			methodReturnType = this.methodDecl.returnType().toString();
//		}
		methodReturnType = this.methodDecl.returnType().toString();
       
		code.append("\n");
		if (this.javadoc != null) {
			if (instrumentationType.equals("clientAwareChecking")) {
				if (visibility == ACC_PUBLIC) {
					code.append(this.javadoc.replace("postcondition", "public postcondition"));
				} else if (visibility == ACC_PROTECTED) {
					code.append(this.javadoc.replace("postcondition", "protected postcondition"));
				} else if (visibility == 0L) { //default
					code.append(this.javadoc.replace("postcondition", "default postcondition"));
				} else if (visibility == ACC_PRIVATE) {
					code.append(this.javadoc.replace("postcondition", "private postcondition"));
				}
			}
			else{
				code.append(this.javadoc);
			}
		} else {
			code.append("/** Generated by JML to check assertions. */");
		}
		code.append("\n");
		
		if(isMethodCrosscutSpecChecking){
			if(isXCSPointcutStatic){
				code.append("after (");
			}
			else{
				code.append("after (final ").append(classQualifiedName).append(" ").append("object$rac");
			}
			code.append(this.generateAdviceMethodParameters(parameters, isXCSPointcutStatic, isMethodCrosscutSpecChecking, 
					AspectUtil.getInstance().isAspectAdvice(this.methodDecl)).toString());
			code.append(")").append(" throwing (Throwable rac$e)");
//			code.append(this.generateMethodThrowsList(exceptionsInSignalsClauses));
			code.append(" :");
			code.append("\n");
			if(instrumentationType.equals("clientAwareChecking")){
				code.append("   (").append(xcsPointcut).append(")");
				if(visibility == ACC_PROTECTED){
					code.append(" && ");
					code.append("\n");
					code.append("   (within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*) || within("+classQualifiedName+"+))");
				}
				else if(visibility == 0L){//package
					code.append(" && ");
					code.append("\n");
					code.append("   within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*)");
				}
				else if(visibility == ACC_PRIVATE){
					code.append(" && ");
					code.append("\n");
					code.append("   within("+classQualifiedName+")");
				}
				code.append(" && ");
				code.append("\n");
				code.append("   !within(*..AspectJMLRac_*) && !within(AspectJMLRac_*)");
			}
			else if(instrumentationType.equals("callSite")){
				code.append("   (").append(xcsPointcut).append(")");
				code.append(" && ");
				code.append("\n");
				code.append("   !within(*..AspectJMLRac_*) && !within(AspectJMLRac_*)");
			}
			else{
				code.append("   (").append(xcsPointcut).append(")");	
			}
			if(!isXCSPointcutStatic){
				code.append(" && ");
				code.append("\n");
				code.append("   target(object$rac)");
			}
			else{
				if(!flexibleXCS){
					code.append(" && ");
					code.append("\n");
					code.append("   if(JMLChecker.checkJP(thisJoinPoint.getSignature().getDeclaringType(), \"$\"))".replace("$", classQualifiedName));
				}
			}
		} // end crosscutting spec checking
		else{
			String adviceexecution = "";
			String ifPC = "";
			if(AspectUtil.getInstance().isAspectAdvice(this.methodDecl)){
				adviceexecution = " || (adviceexecution(";
				String methodNameWithParameterTypes = methodDecl.ident()+AspectUtil.generateMethodParameters(methodDecl.parameters(), false).toString();
				ifPC = " && if(JMLChecker.getThisJoinPointPartialMethSig(thisJoinPoint.getSignature().toLongString(), " +
						"thisJoinPoint.getSignature().getDeclaringTypeName()).endsWith(\""+methodNameWithParameterTypes+"\"))))";
				
			}
			if(this.isStatic){
				code.append(" ").append("after (");	
				code.append(this.generateAdviceMethodParameters(parameters, this.methodDecl.isStatic(), isMethodCrosscutSpecChecking, 
						AspectUtil.getInstance().isAspectAdvice(this.methodDecl)).toString());
			}
			else{
				if((instrumentationType.equals("callSite") || instrumentationType.equals("clientAwareChecking")) && (this.methodDecl.isConstructor())){
					code.append("after (");
					code.append(this.generateAdviceMethodParameters(parameters, true, isMethodCrosscutSpecChecking, 
							AspectUtil.getInstance().isAspectAdvice(this.methodDecl)).toString());	
				}
				else{
					code.append("after (final ").append(classQualifiedName).append(" ").append("object$rac");
					code.append(this.generateAdviceMethodParameters(parameters, this.methodDecl.isStatic(), isMethodCrosscutSpecChecking, 
							AspectUtil.getInstance().isAspectAdvice(this.methodDecl)).toString());	
				}
			}
			code.append(")").append(" throwing (Throwable rac$e) :");
			code.append("\n");
			if(this.methodDecl.isConstructor()){
				if(instrumentationType.equals("clientAwareChecking")){
					code.append("   call(").append(methodQualifiedName).append(")");
					if(visibility == ACC_PROTECTED){
						code.append(" && ");
						code.append("\n");
						code.append("   (within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*) || within("+classQualifiedName+"+))");
					}
					else if(visibility == 0L){//package
						code.append(" && ");
						code.append("\n");
						code.append("   within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*)");
					}
					else if(visibility == ACC_PRIVATE){
						code.append(" && ");
						code.append("\n");
						code.append("   within("+classQualifiedName+")");
					}
					code.append(" && ");
					code.append("\n");
					code.append("   !within(*..AspectJMLRac_*) && !within(AspectJMLRac_*)");
				}
				else if(instrumentationType.equals("callSite")){
					code.append("   call(").append(methodQualifiedName).append(")");
					code.append(" && ");
					code.append("\n");
					code.append("   !within(*..AspectJMLRac_*) && !within(AspectJMLRac_*)");
				}
				else{
					code.append("   execution(").append(methodQualifiedName).append(")");
				}
			}
			else if(this.isStatic){
				if(instrumentationType.equals("clientAwareChecking")){
					code.append("   call(").append("static").append(" ").append(methodReturnType).append(" ").append(methodQualifiedName).append("(").append(AspectUtil.generateMethodParametersForAdvicePointcut(parameters).toString()).append(")").append(")");
					if(visibility == ACC_PROTECTED){
						code.append(" && ");
						code.append("\n");
						code.append("   (within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*) || within("+classQualifiedName+"+))");
					}
					else if(visibility == 0L){//package
						code.append(" && ");
						code.append("\n");
						code.append("   within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*)");
					}
					else if(visibility == ACC_PRIVATE){
						code.append(" && ");
						code.append("\n");
						code.append("   within("+classQualifiedName+")");
					}
					code.append(" && ");
					code.append("\n");
					code.append("   !within(*..AspectJMLRac_*) && !within(AspectJMLRac_*)");
				}
				else if(instrumentationType.equals("callSite")){
					code.append("   call(").append("static").append(" ").append(methodReturnType).append(" ").append(methodQualifiedName).append("(").append(AspectUtil.generateMethodParametersForAdvicePointcut(parameters).toString()).append(")").append(")");
					code.append(" && ");
					code.append("\n");
					code.append("   !within(*..AspectJMLRac_*) && !within(AspectJMLRac_*)");
				}
				else{
					code.append("   execution(").append("static").append(" ").append(methodReturnType).append(" ").append(methodQualifiedName).append("(").append(AspectUtil.generateMethodParametersForAdvicePointcut(parameters).toString()).append(")").append(")");	
				}
			}
			else{
				if(instrumentationType.equals("clientAwareChecking")){
					code.append("   (call(").append(methodReturnType).append(" ").append(methodQualifiedName).append("(").append(AspectUtil.generateMethodParametersForAdvicePointcut(parameters).toString()).
					append(")").append(")").append(adviceexecution).append(")").append(ifPC);
					if(visibility == ACC_PROTECTED){
						code.append(" && ");
						code.append("\n");
						code.append("   (within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*) || within("+classQualifiedName+"+))");
					}
					else if(visibility == 0L){//package
						code.append(" && ");
						code.append("\n");
						code.append("   within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*)");
					}
					else if(visibility == ACC_PRIVATE){
						code.append(" && ");
						code.append("\n");
						code.append("   within("+classQualifiedName+")");
					}
					code.append(" && ");
					code.append("\n");
					code.append("   !within(*..AspectJMLRac_*) && !within(AspectJMLRac_*)");
				}
				else if(instrumentationType.equals("callSite")){
					code.append("   (call(").append(methodReturnType).append(" ").append(methodQualifiedName).append("(").append(AspectUtil.generateMethodParametersForAdvicePointcut(parameters).toString()).
					append(")").append(")").append(adviceexecution).append(")").append(ifPC);
					code.append(" && ");
					code.append("\n");
					code.append("   !within(*..AspectJMLRac_*) && !within(AspectJMLRac_*)");
				}
				else{
					code.append("   (execution(").append(methodReturnType).append(" ").append(methodQualifiedName).append("(").append(AspectUtil.generateMethodParametersForAdvicePointcut(parameters).toString()).
					append(")").append(")").append(adviceexecution).append(")").append(ifPC);	
				}
			}
			if(!this.isStatic){
				if(instrumentationType.equals("callSite") || instrumentationType.equals("clientAwareChecking")){
					if(!this.methodDecl.isConstructor()){
						code.append(" && ");
						code.append("target(object$rac)");
					}
				}
				else{
					code.append(" && ");
					code.append("this(object$rac)");
				}
			}
		}
		
		if(!isMethodCrosscutSpecChecking){
			this.generateMethodArgsForTheAdvice(parameters, code, AspectUtil.getInstance().isAspectAdvice(this.methodDecl));
		}
		String result = code.toString();
		return result;
	}
	
//	private String generateMethodThrowsList(CClassType[] exceptions) {
//		StringBuffer code = new StringBuffer("");
//		if(exceptions != null){
//			if(exceptions.length > 0){
//				code.append(" ");
//				code.append("throws ");
//			}
//			for (int i = 0; i < exceptions.length; i++) {
//				if (i != 0)
//					code.append(", ");
//				String ident = exceptions[i].qualifiedName().replace("/", ".").replace("$", ".");
//				code.append(ident);
//			}	
//		}
//		return code.toString();
//	}
	
//	private String generateMethodThrowsList(List<CType> exceptionsInSignalsClauses) {
//		StringBuffer code = new StringBuffer("");
//		if(exceptionsInSignalsClauses.size() > 0){
//			code.append(" ");
//			code.append("throws ");
//			int i = 0;
//			for (Iterator<CType> iterator = exceptionsInSignalsClauses.iterator(); iterator
//					.hasNext();) {
//				if (i != 0)
//					code.append(", ");
//				CType cType = iterator.next();
//				code.append(cType);
//				i++;
//				
//			}
//		}
//		return code.toString();
//	}

	protected String buildAssertionMethodCall(String methodName, String ownerType, 
			JFormalParameter[] parameters, CClassType[] exceptions, String assertionMethodType) 
	{
		final StringBuffer code = new StringBuffer();
		code.append(methodName);

		// formal parameters
		code.append("(");
		if (parameters != null) {
			for (int i = 0; i < parameters.length; i++) {
				if (i != 0)
					code.append(", ");

				parameters[i].accept( new JmlAbstractVisitor() {
					public void	visitJmlFormalParameter(
							JmlFormalParameter self) {
						String ident = self.ident();
						code.append(ident);
					}
				});
			}
		}
		if(assertionMethodType.equals("NormalPostconditionAssertionMethod")){
			if(this.parameters != null && this.parameters.length > 0){
				code.append(", ");	
			}
			code.append(" ").append("rac$result");
		}
		if(assertionMethodType.equals("ExceptionalPostconditionAssertionMethod")){
			if(this.parameters != null && this.parameters.length > 0){
				code.append(", ");	
			}
			code.append(" ").append("rac$e");
		}
		
		if(Main.aRacOptions.clientAwareChecking()){
			if(this.parameters != null && this.parameters.length > 0){
				code.append(", ");	
			}
			code.append("object$rac");
		}

		code.append(")");

		return code.toString();
	}

	/**
	 * @return
	 */
	protected String generateErrorMessage(String context, String tokenReference, String errorType, boolean isMethodCrosscutSpecChecking) {
		String result = "";

		if ((errorType.equals("Precondition_Error"))
				|| (errorType.equals("Precondition_EvaluationError"))
				|| (errorType.equals("XPost_Error"))) {
			result = this.generatePreAndXPostErrorMessage(context,
					tokenReference, errorType, isMethodCrosscutSpecChecking);
		} else if ((errorType.equals("NormalPostconditionError"))
				|| (errorType.equals("NPost_EvaluationError"))) {
			result = this.generateNPostErrorMessage(context, tokenReference,
					errorType, isMethodCrosscutSpecChecking);
		}
		return result;
	}

	/**
	 * @param context
	 * @param errorType
	 * @return
	 */
	private String generatePreAndXPostErrorMessage(String context,
			String tokenReference, String errorType, boolean isMethodCrosscutSpecChecking) {
//		List inheritedMethsWithPreITD = getMethsWithPreITDChecking();
//		String inheritedPreFromSuperTypes = "";
		String methodDescription = "";
		if(isMethodCrosscutSpecChecking){
			methodDescription = "by method \"+thisJoinPoint.getSignature().getDeclaringTypeName()+\".\"+thisJoinPoint.getSignature().getName()+\"";
		}
		else{
			methodDescription = "by method "+this.methodDecl.getMethod().getJavaName();
		}
		String evalErrorDescription = AssertionMethod.SPEC_EVAL_ERROR_MSG;
		final String SPEC_ERROR_MSG = AssertionMethod.SPEC_ERROR_MSG;
		final String CODE_ERROR_MSG = AssertionMethod.CODE_ERROR_MSG;
		final String NONE = "";
		String result = "";

		// resolving inherited precondition...
//		inheritedPreFromSuperTypes = this.getInheritedPreFromSuperTypes(
//				inheritedMethsWithPreITD, inheritedPreFromSuperTypes);

		if(this.methodDecl.body()!= null){
			JStatement js[] = null;
			
			try {
				js = this.methodDecl.body().body();
				
				if(context.startsWith("\\")){
					context = "+\""+context;
				}
				String bodyCodeReferenceWithoutContext = 
					( js.length == 0 ? "\"" :  
						", line "+
						+ js[0].getTokenReference().line() + " ("
						+ this.typeDecl.getCClass().getJavaName() + ".java:"
						+ js[0].getTokenReference().line() + ")\"");
				String bodyCodeReferenceWithContext = bodyCodeReferenceWithoutContext + (context.length()> 0 ? "+\", when \\n\""+context : "");
				
				String bodyCodeReferenceWithoutContextForXPost = 
					( js.length == 0 ? "\"" :  
						", line "+
						+ js[js.length-1].getTokenReference().line() + " ("
						+ this.typeDecl.getCClass().getJavaName() + ".java:"
						+ js[js.length-1].getTokenReference().line() + ")\"");
				
				String bodyCodeReferenceWithContextForXPost = bodyCodeReferenceWithoutContextForXPost + (context.length()> 0 ? "+\", when \\n\""+context : "");
				
				if (errorType.equals("Precondition_EvaluationError")){
					//result = evalErrorDescription + this.typeDecl.ident() + ".java\\\"" + bodyCodeReferenceWithoutContext;
					if(!tokenReference.equals(NONE)){
						result = evalErrorDescription + this.typeDecl.getCClass().getJavaName() + ".java\\\"" + " " + methodDescription + SPEC_ERROR_MSG + tokenReference + (context.length()> 0 ? "\"+\", when \\n\""+context+"+\"" : "") + "\\nCaused by: \"";	
					}
					else{
						result = evalErrorDescription + this.typeDecl.getCClass().getJavaName() + ".java\\\"\\nCaused by: \"";
					}

				} else if (errorType.equals("Precondition_Error")){
					String tmp = "";

					if(!tokenReference.equals(NONE)){
						tmp = ", " + tokenReference + ", and \\n" + methodDescription + CODE_ERROR_MSG + "File \\\""+ this.typeDecl.getCClass().getJavaName() + ".java\\\"";
						result =  "\"" + methodDescription + SPEC_ERROR_MSG + "File \\\"" + this.typeDecl.getCClass().getJavaName() + ".java\\\"";
						result += tmp + bodyCodeReferenceWithContext;
					}
					else{
						tmp = "\"" + methodDescription + CODE_ERROR_MSG + "File \\\""+ this.typeDecl.getCClass().getJavaName() + ".java\\\"";
						result += tmp + bodyCodeReferenceWithContext;
						if((js.length == 0) && (context.length() > 0)){
							result = result + "+ \", when \\n\""+context;
						}
					}

				} else if (errorType.equals("XPost_Error")){
					String tmp = "";
					if(!tokenReference.equals(NONE)){
						tmp = ", " + tokenReference + ", and \\n" + methodDescription + CODE_ERROR_MSG + "File \\\""+ this.typeDecl.getCClass().getJavaName() + ".java\\\"";
						result =  "\"" + methodDescription + SPEC_ERROR_MSG + "File \\\"" + this.typeDecl.getCClass().getJavaName() + ".java\\\"";
						result += tmp + bodyCodeReferenceWithContextForXPost;
					}
					else{
						tmp = "\"" + methodDescription + CODE_ERROR_MSG + "File \\\""+ this.typeDecl.getCClass().getJavaName() + ".java\\\"";
						result += tmp + bodyCodeReferenceWithContextForXPost;
					}

					if((js.length == 0) && (context.length() > 0)){
						result = result + "+ \", when \\n\""+context;
					}
				}
			} catch (NullPointerException e) {
				if(context.startsWith("\\")){
					context = "+\""+context;
				}
				String bodyCodeReferenceWithoutContext = "\"";
				String bodyCodeReferenceWithContext = bodyCodeReferenceWithoutContext + (context.length()> 0 ? "+\", when \\n\""+context : "");
				
				String bodyCodeReferenceWithoutContextForXPost = "\"";
				
				String bodyCodeReferenceWithContextForXPost = bodyCodeReferenceWithoutContextForXPost + (context.length()> 0 ? "+\", when \\n\""+context : "");
				
				if (errorType.equals("Precondition_EvaluationError")){
					//result = evalErrorDescription + this.typeDecl.ident() + ".java\\\"" + bodyCodeReferenceWithoutContext;
					if(!tokenReference.equals(NONE)){
						result = evalErrorDescription + this.typeDecl.getCClass().getJavaName() + ".java\\\"" + " " + methodDescription + SPEC_ERROR_MSG + tokenReference + (context.length()> 0 ? "\"+\", when \\n\""+context+"+\"" : "") + "\\nCaused by: \"";	
					}
					else{
						result = evalErrorDescription + this.typeDecl.getCClass().getJavaName() + ".java\\\"\\nCaused by: \"";
					}

				} else if (errorType.equals("Precondition_Error")){
					String tmp = "";
					if(!tokenReference.equals(NONE)){
						tmp = ", " + tokenReference + ", and \\n" + methodDescription + CODE_ERROR_MSG + "File \\\""+ this.typeDecl.getCClass().getJavaName() + ".java\\\"";
						result =  "\"" + methodDescription + SPEC_ERROR_MSG + "File \\\"" + this.typeDecl.getCClass().getJavaName() + ".java\\\"";
						result += tmp + bodyCodeReferenceWithContext;
					}
					else{
						tmp = "\"" + methodDescription + CODE_ERROR_MSG + "File \\\""+ this.typeDecl.getCClass().getJavaName() + ".java\\\"";
						result += tmp + bodyCodeReferenceWithContext;
						if(((context.length() > 0))){
							result = result + "+ \", when \\n\""+context;
						}
					}
				} else if (errorType.equals("XPost_Error")){
					String tmp = "";
					if(!tokenReference.equals(NONE)){
						tmp = ", " + tokenReference + ", and \\n" + methodDescription + CODE_ERROR_MSG + "File \\\""+ this.typeDecl.getCClass().getJavaName() + ".java\\\"";
						result =  "\"" + methodDescription + SPEC_ERROR_MSG + "File \\\"" + this.typeDecl.getCClass().getJavaName() + ".java\\\"";
						result += tmp + bodyCodeReferenceWithContextForXPost;
					}
					else{
						tmp = "\"" + methodDescription + CODE_ERROR_MSG + "File \\\""+ this.typeDecl.getCClass().getJavaName() + ".java\\\"";
						result += tmp + bodyCodeReferenceWithContextForXPost;
					}

					if((context.length() > 0)){
						// never throw an exception... so it is not necessary to create context msg... (henrique rebelo)
						result = "\"\"";
					}
				}
			}
		}
		else{
			// If has no body, which occurs with methods inside Interfaces, the only possibility here is "XPost_Error"
			// because for preconditions, only intertype declarations are generated... (there is no precondition_error code generated)
			
			if (errorType.endsWith("_Error")){
				String tmp = "";
				result = "\""
						+ (!tokenReference.equals(NONE) ? ", " + tokenReference
								+ ", and \\n" + methodDescription
								+ CODE_ERROR_MSG + "File \\\""
								+ this.typeDecl.getCClass().getJavaName()
								+ ".java\\\"" : "")
						+ methodDescription
						+ SPEC_ERROR_MSG
						+ "File \\\""
						+ this.typeDecl.getCClass().getJavaName()
						+ ".java\\\""
						+ (context.length() > 0 ? ", when \\n" + context : "\"");
			}
			
			if (errorType.equals("Precondition_EvaluationError")){
				if(!tokenReference.equals(NONE)){
					result = evalErrorDescription + this.typeDecl.getCClass().getJavaName() + ".java\\\"" + " " + methodDescription + SPEC_ERROR_MSG + tokenReference + (context.length()> 0 ? "\"+\", when \\n\"+\""+context+"+\"" : "") + "\\nCaused by: \"";	
				}
				else{
					result = evalErrorDescription + this.typeDecl.getCClass().getJavaName() + ".java\\\"\\nCaused by: \"";
				}

			}
		}

		return result;
	}
	
	

	private String generateNPostErrorMessage(String context,
			String tokenReference, String errorType, boolean isMethodCrosscutSpecChecking) {
		String methodDescription = "";
		if(isMethodCrosscutSpecChecking){
			methodDescription = "by method \"+thisJoinPoint.getSignature().getDeclaringTypeName()+\".\"+thisJoinPoint.getSignature().getName()+\"";
		}
		else{
			methodDescription = "by method "+this.methodDecl.getMethod().getJavaName();
		}
		String evalErrorDescription = AssertionMethod.SPEC_EVAL_ERROR_MSG;
		final String SPEC_ERROR_MSG = AssertionMethod.SPEC_ERROR_MSG;
		final String CODE_ERROR_MSG = AssertionMethod.CODE_ERROR_MSG;
		final String NONE = "";
		String result = "";

		if(this.methodDecl.body()!= null){
			JStatement js[] = null;
			
			try {
				js = this.methodDecl.body().body();
				if(context.startsWith("\\")){
					context = "+\""+context;
				}
				String bodyCodeReferenceWithoutContext = 
					( js.length == 0 ? "\"" :  
						", line "+
						+ js[js.length-1].getTokenReference().line() + " ("
						+ this.typeDecl.getCClass().getJavaName() + ".java:"
						+ js[js.length-1].getTokenReference().line() + ")\"");
				String bodyCodeReferenceWithContext = bodyCodeReferenceWithoutContext + (context.length()> 0 ? "+\", when \\n\""+context : "");

				if (errorType.equals("NPost_EvaluationError")){
					if(!tokenReference.equals(NONE)){
						result = evalErrorDescription + this.typeDecl.getCClass().getJavaName() + ".java\\\"" + " " + methodDescription + SPEC_ERROR_MSG + tokenReference + (context.length()> 0 ? "\"+\", when \\n\""+context+"+\"" : "") + "\\nCaused by: \"";	
					}
					else{
						result = evalErrorDescription + this.typeDecl.getCClass().getJavaName() + ".java\\\"\\nCaused by: \"";
					}
				}
				else{
					String tmp = "";
					if(!tokenReference.equals(NONE)){
						tmp = ", " + tokenReference + ", and \\n" + methodDescription + CODE_ERROR_MSG + "File \\\""+ this.typeDecl.getCClass().getJavaName() + ".java\\\"";
						result =  "\"" + methodDescription + SPEC_ERROR_MSG + "File \\\"" + this.typeDecl.getCClass().getJavaName() + ".java\\\"";
						result += tmp + bodyCodeReferenceWithContext;
					}
					else{
						tmp = "\"" + methodDescription + CODE_ERROR_MSG + "File \\\""+ this.typeDecl.getCClass().getJavaName() + ".java\\\"";
						result += tmp + bodyCodeReferenceWithContext;
						if((js.length == 0) && (context.length() > 0)){
							result = result + "+\", when \\n\""+context;
						}
					}
				}
			} catch (NullPointerException e) {
				if(context.startsWith("\\")){
					context = "+\""+context;
				}
				String bodyCodeReferenceWithoutContext = "\"";
					
				String bodyCodeReferenceWithContext = bodyCodeReferenceWithoutContext + (context.length()> 0 ? "+\", when \\n\""+context : "");

				if (errorType.equals("NPost_EvaluationError")){
					if(!tokenReference.equals(NONE)){
						result = evalErrorDescription + this.typeDecl.getCClass().getJavaName() + ".java\\\"" + " " + methodDescription + SPEC_ERROR_MSG + tokenReference + (context.length()> 0 ? "\"+\", when \\n\""+context+"+\"" : "") + "\\nCaused by: \"";	
					}
					else{
						result = evalErrorDescription + this.typeDecl.getCClass().getJavaName() + ".java\\\"\\nCaused by: \"";
					}
				}
				else{
					String tmp = "";
					if(!tokenReference.equals(NONE)){
						tmp = ", " + tokenReference + ", and \\n" + methodDescription + CODE_ERROR_MSG + "File \\\""+ this.typeDecl.getCClass().getJavaName() + ".java\\\"";
						result =  "\"" + methodDescription + SPEC_ERROR_MSG + "File \\\"" + this.typeDecl.getCClass().getJavaName() + ".java\\\"";
						result += tmp + bodyCodeReferenceWithContext;
					}
					else{
						tmp = "\"" + methodDescription + CODE_ERROR_MSG + "File \\\""+ this.typeDecl.getCClass().getJavaName() + ".java\\\"";
						result += tmp + bodyCodeReferenceWithContext;
						if((context.length() > 0)){
							result = result + "+\", when \\n\""+context;
						}
					}
				}
			}
		}
		else{
			if (errorType.equals("NPost_EvaluationError")){
				if(!tokenReference.equals(NONE)){
					result = evalErrorDescription + this.typeDecl.getCClass().getJavaName() + ".java\\\"" + " " + methodDescription + SPEC_ERROR_MSG + tokenReference + (context.length()> 0 ? "\"+\", when \\n"+context+"+\"" : "") + "\\nCaused by: \"";	
				}
				else{
					result = evalErrorDescription + this.typeDecl.getCClass().getJavaName()+ ".java\\\"\\nCaused by: \"";
				}
			}
			else{
				if(!tokenReference.equals(NONE)){
					result = "\"" + methodDescription + SPEC_ERROR_MSG + "File \\\"" + this.typeDecl.getCClass().getJavaName() + ".java\\\"" + ", "+tokenReference + (context.length()> 0 ? ", when \\n"+context : "\"");
				}
				else{
					result = "\"" + methodDescription + " at \\n" + "File \\\"" + this.typeDecl.getCClass().getJavaName() + ".java\\\"" + (context.length()> 0 ? ", when \\n"+context : "\"");
				}	
			}
		}
		return result;
	}

//	private String getInheritedPreFromSuperTypes(List inheritedMethsWithPreITD,
//			String inheritedPreFromSuperTypes) {
//
//		for (Iterator iterator = inheritedMethsWithPreITD.iterator(); iterator.hasNext();) {
//			CMethod currentMethod = (CMethod) iterator.next();
//			String type = currentMethod.getJavaName();
//			type = type.replace("."+currentMethod.ident(), "");
//			inheritedPreFromSuperTypes += "File \\\""+type+".java\\\", which is inherited\\n"; 				
//		}
//		return inheritedPreFromSuperTypes;
//	}

	/**
	 * @param code
	 * @param classQualifiedName
	 */
	private String buildPreconditionAdviceHeader(String classQualifiedName, String methodQualifiedName, String instrumentationType, 
			long visibility, String methodQualifiedNameJP, boolean isMethodCrosscutSpecChecking, 
			String xcsPointcut, boolean isXCSPointcutStatic, boolean flexibleXCS) {
		StringBuffer code = new StringBuffer("");
		String packageQualifiedName = this.methodDecl.getMethod().owner().packageName().replace("/", ".");
//		JavaMethod jm = AspectUtil.getInstance().getCorrespondingJavaMethodThroughJMLMethod(this.methodDecl.getMethod().owner().getJavaName(), this.methodDecl);
		String methodReturnType = "";
//		if((jm != null) && (!(this.methodDecl.isConstructor()))){
//			if(jm.getReturnType().toString().equals(this.methodDecl.returnType().toString())){
//				methodReturnType = this.methodDecl.returnType().toString();
//			}
//			else{
//				methodReturnType = jm.getReturnType().toString();
//			}
//		}
//		else{
//			methodReturnType = this.methodDecl.returnType().toString();
//		}
		methodReturnType = this.methodDecl.returnType().toString();
		
		code.append("\n");
		if (this.javadoc != null) {
			if (instrumentationType.equals("clientAwareChecking")) {
				if (visibility == ACC_PUBLIC) {
					code.append(this.javadoc.replace("precondition", "public precondition"));
				} else if (visibility == ACC_PROTECTED) {
					code.append(this.javadoc.replace("precondition", "protected precondition"));
				} else if (visibility == 0L) { //default
					code.append(this.javadoc.replace("precondition", "default precondition"));
				} else if (visibility == ACC_PRIVATE) {
					code.append(this.javadoc.replace("precondition", "private precondition"));
				}
			}
			else{
				code.append(this.javadoc);
			}
		} else {
			code.append("/** Generated by JML to check assertions. */");
		}
		code.append("\n");
		if(isMethodCrosscutSpecChecking){
			if(isXCSPointcutStatic){
				code.append("before (");
			}
			else{
				code.append("before (final ").append(classQualifiedName).append(" ").append("object$rac");
			}
			code.append(this.generateAdviceMethodParameters(this.parameters, isXCSPointcutStatic, isMethodCrosscutSpecChecking, 
					AspectUtil.getInstance().isAspectAdvice(this.methodDecl)).toString());
			code.append(")").append(" :");
			code.append("\n");
			if(instrumentationType.equals("clientAwareChecking")){
				code.append("   (").append(xcsPointcut).append(")").append(" && ");
				if(visibility == ACC_PROTECTED){
					code.append("\n");
					code.append("   (within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*) || within("+classQualifiedName+"+))");
					code.append(" && ");
				}
				else if(visibility == 0L){//package
					code.append("\n");
					code.append("   within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*)");
					code.append(" && ");
				}
				else if(visibility == ACC_PRIVATE){
					code.append("\n");
					code.append("   within("+classQualifiedName+")");
					code.append(" && ");
				}
				code.append("   !within(*..AspectJMLRac_*) && !within(AspectJMLRac_*)");
			}
			else if(instrumentationType.equals("callSite")){
				code.append("   (").append(xcsPointcut).append(")").append(" && ");	
				code.append("   !within(*..AspectJMLRac_*) && !within(AspectJMLRac_*)");
			}
			else{
				code.append("   (").append(xcsPointcut).append(")");		
			}
			if(!isXCSPointcutStatic){
				code.append(" && ");
				code.append("\n");
				code.append("   target(object$rac)");
			}
			else{
				if(!flexibleXCS){
					code.append(" && ");
					code.append("\n");
					code.append("   if(JMLChecker.checkJP(thisJoinPoint.getSignature().getDeclaringType(), \"$\"))".replace("$", classQualifiedName));
				}
			}
			
			// appending pointcut for utilprecondition checkingXCS --- [[[hemr]]]
			String pc = "";
			if(instrumentationType.equals("clientAwareChecking")){
				pc = isXCSPointcutStatic? "("+xcsPointcut+")" : "("+xcsPointcut+") && target("+classQualifiedName+")";
			}
			else if(instrumentationType.equals("callSite")){
				pc = isXCSPointcutStatic? "("+xcsPointcut+")" : "("+xcsPointcut+") && target("+classQualifiedName+")";
			}
			else{
				pc = isXCSPointcutStatic? "("+xcsPointcut+")" : "("+xcsPointcut+") && target("+classQualifiedName+")";
			}
			AspectUtil.getInstance().appendPreconditionPointcut(pc);
			
		} // end isMethodCrosscutingSpecChecking
		else{
			String adviceexecution = "";
			String ifPC = "";
			if(AspectUtil.getInstance().isAspectAdvice(this.methodDecl)){
				adviceexecution = " || (adviceexecution(";
				String methodNameWithParameterTypes = methodDecl.ident()+AspectUtil.generateMethodParameters(methodDecl.parameters(), false).toString();
				ifPC = " && if(JMLChecker.getThisJoinPointPartialMethSig(thisJoinPoint.getSignature().toLongString(), " +
						"thisJoinPoint.getSignature().getDeclaringTypeName()).endsWith(\""+methodNameWithParameterTypes+"\"))))";
				
			}
			if(isStatic){
				code.append("before (");	
				code.append(this.generateAdviceMethodParameters(this.parameters, this.methodDecl.isStatic(), isMethodCrosscutSpecChecking, 
						AspectUtil.getInstance().isAspectAdvice(this.methodDecl)).toString());
			}
			else{
				if(instrumentationType.equals("callSite") || instrumentationType.equals("clientAwareChecking")){
					if(this.methodDecl.isConstructor()){
						code.append("before (");
						code.append(this.generateAdviceMethodParameters(this.parameters, true, isMethodCrosscutSpecChecking, 
								AspectUtil.getInstance().isAspectAdvice(this.methodDecl)).toString());
					}  
					else{
						code.append("before (final ").append(classQualifiedName).append(" ").append("object$rac");
						code.append(this.generateAdviceMethodParameters(this.parameters, this.methodDecl.isStatic(), isMethodCrosscutSpecChecking, 
								AspectUtil.getInstance().isAspectAdvice(this.methodDecl)).toString());
					}	
				}
				else{
					code.append("before (final ").append(classQualifiedName).append(" ").append("object$rac");
					code.append(this.generateAdviceMethodParameters(this.parameters, this.methodDecl.isStatic(), isMethodCrosscutSpecChecking, 
							AspectUtil.getInstance().isAspectAdvice(this.methodDecl)).toString());
				}	
			}
			code.append(")").append(" :");
			code.append("\n");
			if(this.methodDecl.isConstructor()){
				if(instrumentationType.equals("clientAwareChecking")){
					code.append("   call(").append(methodQualifiedName).append(")");
					if(visibility == ACC_PROTECTED){
						code.append(" && ");
						code.append("\n");
						code.append("   (within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*) || within("+classQualifiedName+"+))");
					}
					else if(visibility == 0L){//package
						code.append(" && ");
						code.append("\n");
						code.append("   within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*)");
					}
					else if(visibility == ACC_PRIVATE){
						code.append(" && ");
						code.append("\n");
						code.append("   within("+classQualifiedName+")");
					}
					code.append(" && ");
					code.append("\n");
					code.append("   !within(*..AspectJMLRac_*) && !within(AspectJMLRac_*)");
				}
				else if(instrumentationType.equals("callSite")){
					code.append("   call(").append(methodQualifiedName).append(")").append(" && ");
					code.append("\n");
					code.append("   !within(*..AspectJMLRac_*) && !within(AspectJMLRac_*)");
				}
				else{
					code.append("   execution(").append(methodQualifiedName).append(")").append(" && ");
					code.append("\n");
					code.append("   this(object$rac)");
				}
			}
			else if(this.isStatic){
				if(instrumentationType.equals("clientAwareChecking")){
					code.append("   call(").append("static").append(" ").append(methodReturnType.toString()).append(" ").append(methodQualifiedName).append("(").
					append(AspectUtil.generateMethodParametersForAdvicePointcut(this.parameters).toString()).append(")").append(")");
					if(visibility == ACC_PROTECTED){
						code.append(" && ");
						code.append("\n");
						code.append("   (within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*) || within("+classQualifiedName+"+))");
					}
					else if(visibility == 0L){//package
						code.append(" && ");
						code.append("\n");
						code.append("   within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*)");
					}
					else if(visibility == ACC_PRIVATE){
						code.append(" && ");
						code.append("\n");
						code.append("   within("+classQualifiedName+")");
					}
					code.append(" && ");
					code.append("\n");
					code.append("   !within(*..AspectJMLRac_*) && !within(AspectJMLRac_*)");
				}
				else if(instrumentationType.equals("callSite")){
					code.append("   call(").append("static").append(" ").append(methodReturnType).append(" ").append(methodQualifiedName).append("(").
					append(AspectUtil.generateMethodParametersForAdvicePointcut(this.parameters).toString()).append(")").append(")");
					code.append(" && ");
					code.append("\n");
					code.append("   !within(*..AspectJMLRac_*) && !within(AspectJMLRac_*)");
				}
				else{
					code.append("   execution(").append("static").append(" ").append(methodReturnType).append(" ").append(methodQualifiedName).append("(").
					append(AspectUtil.generateMethodParametersForAdvicePointcut(this.parameters).toString()).append(")").append(")");	
				}
			}
			else{
				if(instrumentationType.equals("clientAwareChecking")){
					String call = "(call("+methodReturnType+" "+methodQualifiedNameJP+
							"("+AspectUtil.generateMethodParametersForAdvicePointcut(this.parameters).toString()+"))"+adviceexecution+")"+ifPC;
					AspectUtil.getInstance().appendPreconditionPointcut(call);
					code.append("   (call(").append(methodReturnType).append(" ").append(methodQualifiedName).append("(").
					append(AspectUtil.generateMethodParametersForAdvicePointcut(this.parameters).toString()).append(")").append(")").
					append(adviceexecution).append(")").append(ifPC).append(" && ");
					if(visibility == ACC_PROTECTED){
						code.append("\n");
						code.append("   (within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*) || within("+classQualifiedName+"+))");
						code.append(" && ");
					}
					else if(visibility == 0L){//package
						code.append("\n");
						code.append("   within("+(packageQualifiedName.equals("")?"":packageQualifiedName)+"*)");
						code.append(" && ");
					}
					else if(visibility == ACC_PRIVATE){
						code.append("\n");
						code.append("   within("+classQualifiedName+")");
						code.append(" && ");
					}
					code.append("target(object$rac)"); 	
					code.append(" && ");
					code.append("\n");
					code.append("   !within(*..AspectJMLRac_*) && !within(AspectJMLRac_*)");
				}
				else if(instrumentationType.equals("callSite")){
					String call = "(call("+methodReturnType+" "+methodQualifiedNameJP+
							"("+AspectUtil.generateMethodParametersForAdvicePointcut(this.parameters).toString()+"))"+adviceexecution+")"+ifPC;
					AspectUtil.getInstance().appendPreconditionPointcut(call);
					code.append("   (call(").append(methodReturnType).append(" ").append(methodQualifiedName).append("(").
					append(AspectUtil.generateMethodParametersForAdvicePointcut(this.parameters).toString()).append(")").append(")").
					append(adviceexecution).append(")").append(ifPC).append(" && ");	
					code.append("\n");
					code.append("   target(object$rac)"); 	
					code.append(" && ");
					code.append("\n");
					code.append("   !within(*..AspectJMLRac_*) && !within(AspectJMLRac_*)");
				}
				else{
					String execution = "(execution("+methodReturnType+" "+methodQualifiedNameJP+
							"("+AspectUtil.generateMethodParametersForAdvicePointcut(this.parameters).toString()+"))"+adviceexecution+")"+ifPC;
					AspectUtil.getInstance().appendPreconditionPointcut(execution);
					code.append("   (execution(").append(methodReturnType).append(" ").append(methodQualifiedName).append("(").
					append(AspectUtil.generateMethodParametersForAdvicePointcut(this.parameters).toString()).append(")").append(")").
					append(adviceexecution).append(")").append(ifPC).append(" && ");	
					code.append("\n");
					code.append("   this(object$rac)"); 
				}
			}
		}
		if(!isMethodCrosscutSpecChecking){
			this.generateMethodArgsForTheAdvice(parameters, code, AspectUtil.getInstance().isAspectAdvice(this.methodDecl));
		}
		String result = code.toString();
		return result;
	}

	/**
	 * @param parameters
	 * @param isStatic
	 * @param code
	 */
	private StringBuffer generateAdviceMethodParameters(JFormalParameter[] parameters,
			boolean isStatic, final boolean isMethodCrosscutSpecChecking, final boolean isAspectAdvice) {
		final StringBuffer code = new StringBuffer("");
		if (parameters != null) {
			if(parameters.length > 0) {
				if((this.methodDecl.isConstructor() && this.methodDecl.getMethod().owner().isInnerClass()) && !(isAspectAdvice)){
					if(!isStatic && parameters.length > 1)
						code.append(", ");
					for (int i = 1; i < parameters.length; i++) {
						if (i != 1)
							code.append(", ");
						parameters[i].accept( new JmlAbstractVisitor() {
							public void	visitJmlFormalParameter(
									JmlFormalParameter self) {
								CType actualType = self.getType();
								String ident = self.ident();
//								if(AspectUtil.replaceDollaSymbol(TransUtils.toString(actualType)).equals("org.aspectj.lang.JoinPoint")){
//									if(ident.equals(VN_ASPECTJ_THISJOINPOINT)){
//										JmlRacGenerator.fail(self.getTokenReference(), JmlMessages.INVALID_JOINPOINT_ARG_IDENT, self.ident());
//										System.exit(0);
//									}
//								}
								code.append("final ");
								code.append(AspectUtil.replaceDollaSymbol(TransUtils.toString(actualType)));
								code.append(" ");
								code.append(ident);
							}
						});
					}
				}
				else{
					if(!isStatic){
						if(isMethodCrosscutSpecChecking || isAspectAdvice){
							if(parameters.length == 1) {
								if(!TransUtils.toString(parameters[0].getType()).equals("org.aspectj.lang.JoinPoint")){
									code.append(", ");
								}
							}
							else {
								code.append(", ");
							}
						}
						else{
							code.append(", ");
						}
					}
					for (int i = 0; i < parameters.length; i++) {
						if (i != 0)
							code.append(", ");
						parameters[i].accept( new JmlAbstractVisitor() {
							public void	visitJmlFormalParameter(
									JmlFormalParameter self) {
								CType actualType = self.getType();
								String ident = self.ident();
								if(isMethodCrosscutSpecChecking || isAspectAdvice){
									if(AspectUtil.replaceDollaSymbol(TransUtils.toString(actualType)).equals("org.aspectj.lang.JoinPoint")){
										if(code.length()-2 >= 0){
											code.delete(code.length()-2, code.length()-1); // removing the comma...
										}
										if(!ident.equals(VN_ASPECTJ_THISJOINPOINT)){
											JmlRacGenerator.fail(self.getTokenReference(), JmlMessages.INVALID_JOINPOINT_ARG_IDENT_XCS, self.ident());
											System.exit(0);
										}
									}
									else{
										code.append("final ");
										code.append(AspectUtil.replaceDollaSymbol(TransUtils.toString(actualType)));
										code.append(" ");
										code.append(ident);
									}
								}
								else{
//									if(!isAspectAdvice){
//										if(AspectUtil.replaceDollaSymbol(TransUtils.toString(actualType)).equals("org.aspectj.lang.JoinPoint")){
//											if(ident.equals(VN_ASPECTJ_THISJOINPOINT)){
//												JmlRacGenerator.fail(self.getTokenReference(), JmlMessages.INVALID_JOINPOINT_ARG_IDENT, self.ident());
//												System.exit(0);
//											}
//										}
//									}
									if(AspectUtil.replaceDollaSymbol(TransUtils.toString(actualType)).equals("org.aspectj.lang.JoinPoint")){
										if(ident.equals(VN_ASPECTJ_THISJOINPOINT)){
											if(code.length()-2 >= 0){
												code.delete(code.length()-2, code.length()-1); // removing the comma...
											}
										}
										else{
											code.append("final ");
											code.append(AspectUtil.replaceDollaSymbol(TransUtils.toString(actualType)));
											code.append(" ");
											code.append(ident);
										}
									}
									else{
										code.append("final ");
										code.append(AspectUtil.replaceDollaSymbol(TransUtils.toString(actualType)));
										code.append(" ");
										code.append(ident);
									}
								}
							}
						});	
					}
				}
			}
		}
		if(code.length() > 0){
			// workaround to fix the ", ," if wrongly generated for params --- [[[hemr]]]
			if(code.toString().indexOf("") != -1){
				String advParam = code.toString().replace(", ,", ",");
				code.delete(0, code.length());
				code.append(advParam);	
			}
		}
		return code;
	}
	
	private boolean hasThisJPSpecialParamDecl(JFormalParameter[] parameters){
		boolean ret = false;
		for (int i = 0; i < parameters.length; i++) {
			if(AspectUtil.replaceDollaSymbol(TransUtils.toString(parameters[i].getType())).equals("org.aspectj.lang.JoinPoint")){
				ret = true;
				break;
			}
		}
		return ret;
	}
	
	private boolean hasProceedingJPSpecialParamDecl(JFormalParameter[] parameters){
		boolean ret = false;
		for (int i = 0; i < parameters.length; i++) {
			if(AspectUtil.replaceDollaSymbol(TransUtils.toString(parameters[i].getType())).equals("org.aspectj.lang.ProceedingJoinPoint")){
				ret = true;
				break;
			}
		}
		return ret;
	}
	
	private String getProceedingJPVarDeclName(JFormalParameter[] parameters){
		String ret = "pjp";
		for (int i = 0; i < parameters.length; i++) {
			if(AspectUtil.replaceDollaSymbol(TransUtils.toString(parameters[i].getType())).equals("org.aspectj.lang.ProceedingJoinPoint")){
				ret = parameters[i].ident();
				break;
			}
		}
		return ret;
	}
	
	private String getThisJPVarDeclName(JFormalParameter[] parameters){
		String ret = "thisJP";
		for (int i = 0; i < parameters.length; i++) {
			if(AspectUtil.replaceDollaSymbol(TransUtils.toString(parameters[i].getType())).equals("org.aspectj.lang.JoinPoint")){
				ret = parameters[i].ident();
				break;
			}
		}
		return ret;
	}
	
	private StringBuffer generateAdviceMethodParametersForAtAspectJ(JFormalParameter[] parameters,
			boolean isStatic, final boolean isMethodCrosscutSpecChecking, final boolean isAspectAdvice) {
		final StringBuffer code = new StringBuffer("");
		if (parameters != null) {
			if(parameters.length > 0) {
				if((this.methodDecl.isConstructor() && this.methodDecl.getMethod().owner().isInnerClass()) && !(isAspectAdvice)){
					if(!isStatic && parameters.length > 1)
						code.append(", ");
					for (int i = 1; i < parameters.length; i++) {
						if (i != 1)
							code.append(", ");
						parameters[i].accept( new JmlAbstractVisitor() {
							public void	visitJmlFormalParameter(
									JmlFormalParameter self) {
								CType actualType = self.getType();
								String ident = self.ident();
//								this should be commented for AspectJML advice or pointcut... [[[hemr]]]
//								if(AspectUtil.replaceDollaSymbol(TransUtils.toString(actualType)).equals("org.aspectj.lang.JoinPoint")){
//									if(ident.equals(VN_ASPECTJ_THISJOINPOINT)){
//										JmlRacGenerator.fail(self.getTokenReference(), JmlMessages.INVALID_JOINPOINT_ARG_IDENT, self.ident());
//										System.exit(0);
//									}
//								}
								code.append("final ");
								code.append(AspectUtil.replaceDollaSymbol(TransUtils.toString(actualType)));
								code.append(" ");
								code.append(ident);
							}
						});
					}
				}
				else{
					if(!isStatic){
						if(isMethodCrosscutSpecChecking || isAspectAdvice){
//							maybe should be restored at some point --- [[[hemr]]]
//							if(parameters.length == 1) {
//								if(!TransUtils.toString(parameters[0].getType()).equals("org.aspectj.lang.JoinPoint")){
//									code.append(", ");
//								}
//							}
//							else {
								code.append(", ");
//							}
						}
						else{
							code.append(", ");
						}
					}
					for (int i = 0; i < parameters.length; i++) {
						if (i != 0)
							code.append(", ");
						parameters[i].accept( new JmlAbstractVisitor() {
							public void	visitJmlFormalParameter(
									JmlFormalParameter self) {
								CType actualType = self.getType();
								String ident = self.ident();
								if(isMethodCrosscutSpecChecking || isAspectAdvice){
//									if(AspectUtil.replaceDollaSymbol(TransUtils.toString(actualType)).equals("org.aspectj.lang.JoinPoint")){
//										if(code.length()-2 >= 0){
//											code.delete(code.length()-2, code.length()-1); // removing the comma...
//										}
//										if(!ident.equals(VN_ASPECTJ_THISJOINPOINT)){
//											JmlRacGenerator.fail(self.getTokenReference(), JmlMessages.INVALID_JOINPOINT_ARG_IDENT_XCS, self.ident());
//											System.exit(0);
//										}
//									}
//									else{
										code.append("final ");
										code.append(AspectUtil.replaceDollaSymbol(TransUtils.toString(actualType)));
										code.append(" ");
//										if(ident.equals(VN_ASPECTJ_THISJOINPOINT)){
//											code.append(ident+"_ajml");
//										}
//										else{
											code.append(ident);
//										}
//									}
								}
//								else{
//									if(!isAspectAdvice){
//										if(AspectUtil.replaceDollaSymbol(TransUtils.toString(actualType)).equals("org.aspectj.lang.JoinPoint")){
//											if(ident.equals(VN_ASPECTJ_THISJOINPOINT)){
//												JmlRacGenerator.fail(self.getTokenReference(), JmlMessages.INVALID_JOINPOINT_ARG_IDENT, self.ident());
//												System.exit(0);
//											}
//										}
//									}
//									code.append("final ");
//									code.append(AspectUtil.replaceDollaSymbol(TransUtils.toString(actualType)));
//									code.append(" ");
//									code.append(ident);
//								}
							}
						});	
					}
				}
			}
		}
		if(code.length() > 0){
			// workaround to fix the ", ," if wrongly generated for params --- [[[hemr]]]
			if(code.toString().indexOf("") != -1){
				String advParam = code.toString().replace(", ,", ",");
				code.delete(0, code.length());
				code.append(advParam);	
			}
		}
		return code;
	}

	/**
	 * @param parameters
	 * @param code
	 */
	private void generateMethodArgsForTheAdvice(JFormalParameter[] parameters,
			final StringBuffer code, final boolean isAspectAdvice) {
		if (parameters != null) {
			
//			if(this.methodDecl.getMethod().owner().isInnerClass()){
//				if(parameters.length > 1)
//					code.append(" && ").append("args(");
//			}
//			else{
				if(parameters.length > 0)
					code.append(" && ").append("args(");
//			}
			
			if(this.methodDecl.isConstructor() && this.methodDecl.getMethod().owner().isInnerClass()){
				for (int i = 1; i < parameters.length; i++) {
					if (i != 1)
						code.append(", ");
					parameters[i].accept( new JmlAbstractVisitor() {
						public void	visitJmlFormalParameter(
								JmlFormalParameter self) {
							String ident = self.ident();
							if(ident.equals("thisJoinPoint")){
								code.append("..");
							}
							else{
								code.append(ident);
							}
							
						}
					});
				}
			}
			else{
				for (int i = 0; i < parameters.length; i++) {
					if (i != 0)
						code.append(", ");
					parameters[i].accept( new JmlAbstractVisitor() {
						public void	visitJmlFormalParameter(
								JmlFormalParameter self) {
							String ident = self.ident();
							if(ident.equals("thisJoinPoint")){
								code.append("..");
							}
							else{
								code.append(ident);
							}
						}
					});	
				}
			}
 
//			if(this.methodDecl.getMethod().owner().isInnerClass()){
//				if(parameters.length > 1)
//					code.append(")");
//			}
//			else{
				if(parameters.length > 0)
					code.append(")");
//			}
		}
	}
	
	private void generateOptionalArgsPC(JFormalParameter[] parameters,
			final StringBuffer code) {
		if (parameters != null) {

			if(parameters.length > 0)
				code.append(" && ").append("args(");


			if(this.methodDecl.isConstructor() && this.methodDecl.getMethod().owner().isInnerClass()){
				for (int i = 1; i < parameters.length; i++) {
					if (i != 1)
						code.append(", ");
					parameters[i].accept( new JmlAbstractVisitor() {
						public void	visitJmlFormalParameter(
								JmlFormalParameter self) {
							String ident = self.ident();
							if(TransUtils.toString(self.getType()).equals("org.aspectj.lang.JoinPoint")){
								code.append("..");
							}
							else if(TransUtils.toString(self.getType()).equals("org.aspectj.lang.ProceedingJoinPoint")){
								code.append("..");
							}
							else{
								code.append(ident);
							}
						}
					});
				}
			}
			else{
				for (int i = 0; i < parameters.length; i++) {
					if (i != 0)
						code.append(", ");
					parameters[i].accept( new JmlAbstractVisitor() {
						public void	visitJmlFormalParameter(
								JmlFormalParameter self) {
							String ident = self.ident();
							if(TransUtils.toString(self.getType()).equals("org.aspectj.lang.JoinPoint")){
								code.append("..");
							}
							else if(TransUtils.toString(self.getType()).equals("org.aspectj.lang.ProceedingJoinPoint")){
								code.append("..");
							}
							else{
								code.append(ident);
							}
						}
					});	
				}
			}

			if(parameters.length > 0){
				code.append(",.."); // used for matching argments in sequence if not declared for all parameters [enhancement] --- [[[hemr]]] 
				code.append(")");
			}
		}
	}
	
	protected HashMap getPreconditions(long visibility){
		if (visibility == Constants.ACC_PUBLIC) {
			return AspectUtil.getInstance().publicPreconditionSpecCases;
		} else if (visibility == Constants.ACC_PROTECTED) {
			return AspectUtil.getInstance().protectedPreconditionSpecCases;
		} else if (visibility == 0L) { //default
			return AspectUtil.getInstance().defaultPreconditionSpecCases;
		} else if (visibility == Constants.ACC_PRIVATE) {
			return AspectUtil.getInstance().privatePreconditionSpecCases;
		}
		else{
			return AspectUtil.getInstance().preconditionSpecCases;
		}
	}
	
	protected HashMap getNormalPostconditions(long visibility){
		if (visibility == Constants.ACC_PUBLIC) {
			return AspectUtil.getInstance().publicNPostconditionSpecCases;
		} else if (visibility == Constants.ACC_PROTECTED) {
			return AspectUtil.getInstance().protectedNPostconditionSpecCases;
		} else if (visibility == 0L) { //default
			return AspectUtil.getInstance().defaultNPostconditionSpecCases;
		} else if (visibility == Constants.ACC_PRIVATE) {
			return AspectUtil.getInstance().privateNPostconditionSpecCases;
		}
		else{
			return AspectUtil.getInstance().nPostconditionSpecCases;
		}
	}
	
	protected HashMap getContextForNPost(long visibility){
		if (visibility == Constants.ACC_PUBLIC) {
			return AspectUtil.getInstance().publicNPostconditionForContextSpecCases;
		} else if (visibility == Constants.ACC_PROTECTED) {
			return AspectUtil.getInstance().protectedNPostconditionForContextSpecCases;
		} else if (visibility == 0L) { //default
			return AspectUtil.getInstance().defaultNPostconditionForContextSpecCases;
		} else if (visibility == Constants.ACC_PRIVATE) {
			return AspectUtil.getInstance().privateNPostconditionForContextSpecCases;
		}
		else{
			return AspectUtil.getInstance().nPostconditionForContextSpecCases;
		}
	}
	
	protected HashMap getContextForXPost(long visibility){
		if (visibility == Constants.ACC_PUBLIC) {
			return AspectUtil.getInstance().publicXPostconditionForContextSpecCases;
		} else if (visibility == Constants.ACC_PROTECTED) {
			return AspectUtil.getInstance().protectedXPostconditionForContextSpecCases;
		} else if (visibility == 0L) { //default
			return AspectUtil.getInstance().defaultXPostconditionForContextSpecCases;
		} else if (visibility == Constants.ACC_PRIVATE) {
			return AspectUtil.getInstance().privateXPostconditionForContextSpecCases;
		}
		else{
			return AspectUtil.getInstance().xPostconditionForContextSpecCases;
		}
	}
	
	protected HashMap getTokenReferenceForNPost(long visibility){
		if (visibility == Constants.ACC_PUBLIC) {
			return AspectUtil.getInstance().publicNPostTokenReferenceSpecCases;
		} else if (visibility == Constants.ACC_PROTECTED) {
			return AspectUtil.getInstance().protectedNPostTokenReferenceSpecCases;
		} else if (visibility == 0L) { //default
			return AspectUtil.getInstance().defaultNPostTokenReferenceSpecCases;
		} else if (visibility == Constants.ACC_PRIVATE) {
			return AspectUtil.getInstance().privateNPostTokenReferenceSpecCases;
		}
		else{
			return AspectUtil.getInstance().nPostTokenReferenceSpecCases;
		}
	}
	
//	protected HashMap getTokenReferenceForXPost(long visibility){
//		if (visibility == Constants.ACC_PUBLIC) {
//			return AspectUtil.getInstance().publicXPostTokenReferenceSpecCases;
//		} else if (visibility == Constants.ACC_PROTECTED) {
//			return AspectUtil.getInstance().protectedXPostTokenReferenceSpecCases;
//		} else if (visibility == 0L) { //default
//			return AspectUtil.getInstance().defaultXPostTokenReferenceSpecCases;
//		} else if (visibility == Constants.ACC_PRIVATE) {
//			return AspectUtil.getInstance().privateXPostTokenReferenceSpecCases;
//		}
//		else{
//			return AspectUtil.getInstance().xPostTokenReferenceSpecCases;
//		}
//	}
	
	protected String buildCallProceed(JFormalParameter[] parameters, String instrumentationType, /*boolean isInnerType,*/ final boolean isMethodCrosscutSpecChecking, boolean isFlexibleXCS) {
		final StringBuffer code = new StringBuffer();
		code.append("proceed");

		// formal parameters
		code.append("(");
		if((!isStatic) && (!isFlexibleXCS)){
			if(instrumentationType.equals("callSite") || instrumentationType.equals("clientAwareChecking")){
				if(!this.methodDecl.isConstructor()){
					code.append("object$rac");
				}
			}
			else{
				code.append("object$rac");
			}
		}
		if (parameters != null) {
				if(parameters.length > 0)
					if(!isStatic){
						if(instrumentationType.equals("callSite") || instrumentationType.equals("clientAwareChecking")){
							if(!this.methodDecl.isConstructor()){
								if(parameters.length == 1) {
									if(!TransUtils.toString(parameters[0].getType()).equals("org.aspectj.lang.JoinPoint")){
										code.append(", ");
									}
								}
								else {
									code.append(", ");
								}
							}
						}
						else{
							if(isMethodCrosscutSpecChecking){
								if(parameters.length == 1) {
									if(!TransUtils.toString(parameters[0].getType()).equals("org.aspectj.lang.JoinPoint")){
										code.append(", ");
									}
								}
								else {
									code.append(", ");
								}
							}
							else{
								code.append(", ");
							}
						}
					}	
				for (int i = 0; i < parameters.length; i++) {
					if (i != 0)
						code.append(", ");

					parameters[i].accept( new JmlAbstractVisitor() {
						public void	visitJmlFormalParameter(
								JmlFormalParameter self) {
							String ident = self.ident();
							if(isMethodCrosscutSpecChecking){
								if(!(AspectUtil.replaceDollaSymbol(TransUtils.toString(self.getType())).equals("org.aspectj.lang.JoinPoint"))){
									code.append(ident);
								}
							}
							else{
								code.append(ident);
							}
						}
					});
				}
			}
//		}

		code.append(")");

		return code.toString();
	}
	
	protected String xPostRethrowStmt(List<CType> exceptionsInSignalsClauses){
		final StringBuffer rethrowStmt = new StringBuffer("");
//		CClassType exceptions[] = this.methodDecl.getExceptions();

		// build exception rethrow statements to be executed if the 
		// exceptional condition be satisfied

		if(!Main.aRacOptions.multipleSpecCaseChecking()){
			rethrowStmt.append("         if(!JMLChecker.hasAnyJMLError) {\n");
			rethrowStmt.append("\t\t\t JMLChecker.rethrowUncheckedException(rac$e);\n");
			rethrowStmt.append("\t\t   }\n");
		}
		else{
			rethrowStmt.append("         JMLChecker.rethrowUncheckedException(rac$e);\n");
		}

		if(exceptionsInSignalsClauses.size() > 0){
			for (Iterator<CType> iterator = exceptionsInSignalsClauses.iterator(); iterator
					.hasNext();) {
				CType cType = iterator.next();
				if ( (!cType.toString().startsWith("java.lang"))){
					rethrowStmt.append("\t\t\t if (" + VN_EXCEPTION);
					rethrowStmt.append(       " instanceof ").append(cType.toString().replace("$", ".")).append(") {\n");
					if(Main.aRacOptions.multipleSpecCaseChecking()){
						rethrowStmt.append("\t\t\t   JMLChecker.resetXPostconditionError();\n");
					}
					rethrowStmt.append("\t\t\t   throw new JMLInternalRuntimeException(("+cType.toString().replace("$", ".")+")" + VN_EXCEPTION + ");\n");
					rethrowStmt.append("\t\t\t }\n");
				}	
			}
		}

		return rethrowStmt.toString();
	}
	
	private String throwStmtsForCC(){
		final StringBuffer throwStmt = new StringBuffer("");
		CClassType exceptions[] = this.methodDecl.getExceptions();

		if(exceptions.length > 0){
			throwStmt.append(" throws ");
			for (int i = 0; i < exceptions.length; i++) {
				CClassType cType = exceptions[i];
				if(i > 0){
					throwStmt.append(", ");
				}
				throwStmt.append(cType.toString().replace("$", "."));
				throwStmt.append(" ");
			}
		}
	
		return throwStmt.toString();
	}
	
	protected String xPostRethrowStmt(){
		final StringBuffer rethrowStmt = new StringBuffer("");
		CClassType exceptions[] = this.methodDecl.getExceptions();

		// build exception rethrow statements to be executed if the 
		// exceptional condition be satisfied

		if(!Main.aRacOptions.multipleSpecCaseChecking()){
			rethrowStmt.append("         if(!JMLChecker.hasAnyJMLError) {\n");
			rethrowStmt.append("\t\t\t JMLChecker.rethrowUncheckedException(rac$e);\n");
			rethrowStmt.append("\t\t   }\n");
		}
		else{
			rethrowStmt.append("         JMLChecker.rethrowUncheckedException(rac$e);\n");
		}

		if(exceptions.length > 0){
			for (int i = 0; i < exceptions.length; i++) {
				CClassType cType = exceptions[i];
				// complementing exception signals set --- [[[hemr]]]
				AspectUtil.getInstance().appendExceptionInSignalsClauseInCompilationUnit(cType);
				if ( (!cType.toString().startsWith("java.lang"))){
					rethrowStmt.append("\t\t\t if (" + VN_EXCEPTION);
					rethrowStmt.append(       " instanceof ").append(cType.toString().replace("$", ".")).append(") {\n");
					if(Main.aRacOptions.multipleSpecCaseChecking()){
						rethrowStmt.append("\t\t\t   JMLChecker.resetXPostconditionError();\n");
					}
					rethrowStmt.append("\t\t\t   throw new JMLInternalRuntimeException(("+cType.toString().replace("$", ".")+")" + VN_EXCEPTION + ");\n");
					rethrowStmt.append("\t\t\t }\n");
				}	
			}
			rethrowStmt.append("\t\t\t else {\n");
			rethrowStmt.append("\t\t\t   throw new JMLSoftException(rac$e);\n");
			rethrowStmt.append("\t\t\t }\n");
		}
		else{
			rethrowStmt.append("\t\t\t throw new JMLSoftException(rac$e);\n");
		}

		return rethrowStmt.toString();
	}
	
	protected String getMultiCatchFromThrowListForAroundAdvice(){
		final StringBuffer code = new StringBuffer("");
		String result = "";
		CClassType exceptions[] = this.methodDecl.getExceptions();
		if(exceptions.length > 0){
			for (int i = 0; i < exceptions.length; i++) {
				CClassType cType = exceptions[i];
				if(code.length() > 0){
					code.append(" | ");
					code.append(cType.toString().replace("$", "."));
				}
				else {
					code.append(cType.toString().replace("$", "."));
				}
				
			}
			result = "catch ("+code.toString()+" | RuntimeException | Error rac$e)";
		}
		else {
			result = "catch (RuntimeException | Error rac$e)";
		}
		return result;
	}
	
	// ----------------------------------------------------------------------
	// DATA MEMBERS
	// ----------------------------------------------------------------------

	/** target method for which assertion method is generated */
	protected final JmlMethodDeclaration methodDecl;

	/** parameters of the target method */
	protected final JFormalParameter[] parameters;

	/** name of the target method */
	protected String name;
}

